#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.io.config`."""

import argparse
import logging
import os

import pytest
import yaml

from orion.core.io.config import Configuration, ConfigurationError


@pytest.fixture
def yaml_path():
    """Create a temporary yaml file and return the path"""
    file_path = './my.yaml'
    with open(file_path, 'w') as f:
        f.write(yaml.dump({'test': 'from my yaml!'}))

    yield file_path

    os.remove(file_path)


@pytest.fixture
def broken_yaml_path():
    """Create a temporary yaml file and return the path"""
    file_path = './my.yaml'
    with open(file_path, 'w') as f:
        f.write(yaml.dump({'coucou': 'from my yaml!'}))

    yield file_path

    os.remove(file_path)


@pytest.fixture
def subdict_yaml_path():
    """Create a temporary yaml file with subdicts and return the path"""
    file_path = './my.yaml'
    with open(file_path, 'w') as f:
        f.write(yaml.dump({'test': {'i am': 'a sub-dict'},
                           'test2': {'test': {'me is': 'sub-conf sub-dict'}}}))

    yield file_path

    os.remove(file_path)


def test_fetch_non_existing_option():
    """Test that access to a non existing key raises ConfigurationError"""
    config = Configuration()
    with pytest.raises(ConfigurationError) as exc:
        config.voici_voila

    assert 'Configuration does not have an attribute \'voici_voila\'.' in str(exc.value)


def test_access_to_config():
    """Test that access to _config returns properly _config.

    This is because getattr() could grasp any key including `_config` and makes
    it impossible to access the later
    """
    assert Configuration()._config == {}


def test_set_subconfig():
    """Test that setting a subconfig works"""
    config = Configuration()
    config.test = Configuration()

    assert isinstance(config.test, Configuration)

    with pytest.raises(ConfigurationError):
        config.test.voici_voila


def test_access_config_without_values():
    """Test that access to config without values raises ConfigurationError"""
    config = Configuration()
    config.add_option('test', option_type=int)
    with pytest.raises(ConfigurationError) as exc:
        config.test

    assert 'Configuration not set and no default provided: test.' in str(exc.value)


def test_set_non_existing_option():
    """Test that setting a non existing option crash"""
    config = Configuration()
    with pytest.raises(TypeError) as exc:
        config.test = 1
    assert "Can only set test as a Configuration, not <class 'int'>" in str(exc.value)


def test_set_subconfig_over_option():
    """Test that overwritting an option with a subconfig is not possible"""
    config = Configuration()
    config.add_option('test', option_type=int)
    config.test = 1
    assert config.test == 1
    with pytest.raises(TypeError) as exc:
        config.test = Configuration()
    assert "Cannot overwrite option test with a configuration" in str(exc.value)


def test_set_int_value():
    """Test that an integer option can have its value set"""
    config = Configuration()
    config.add_option('test', option_type=int)

    with pytest.raises(ConfigurationError) as exc:
        config.test

    config.test = 1
    assert config.test == 1
    config.test = "1"
    assert config.test == 1
    with pytest.raises(TypeError) as exc:
        config.test = "voici_voila"
    assert "<class 'int'> cannot be set to voici_voila with type <class 'str'>" in str(exc.value)


def test_set_real_value():
    """Test that a float option can have its value set"""
    config = Configuration()
    config.add_option('test', option_type=float)

    with pytest.raises(ConfigurationError) as exc:
        config.test

    config.test = 1
    assert config.test == 1.0
    config.test = "1"
    assert config.test == 1.0
    with pytest.raises(TypeError) as exc:
        config.test = "voici_voila"
    assert "<class 'float'> cannot be set to voici_voila with type <class 'str'>" in str(exc.value)


def test_set_str_value():
    """Test that a string option can have its value set"""
    config = Configuration()
    config.add_option('test', option_type=str)

    with pytest.raises(ConfigurationError):
        config.test

    config.test = "1"
    assert config.test == "1"
    config.test = 1
    assert config.test == "1"


def test_set_value_of_subconfig_directly():
    """Test that we can access subconfig and set value directly"""
    config = Configuration()
    config.sub = Configuration()
    config.sub.add_option('test', option_type=str)

    with pytest.raises(ConfigurationError):
        config.test

    config.sub.test = "1"
    assert config.sub.test == "1"
    config.sub.test = 1
    assert config.sub.test == "1"


def test_set_value_like_dict():
    """Test that we can set values like a dictionary"""
    config = Configuration()
    config.add_option('test', option_type=str)

    with pytest.raises(ConfigurationError):
        config.test

    config['test'] = "1"
    assert config.test == "1"
    config['test'] = 1
    assert config.test == "1"


def test_set_subconfig_value_like_dict():
    """Test that we can set values like a dictionary"""
    config = Configuration()
    config.sub = Configuration()
    config.sub.add_option('test', option_type=str)

    with pytest.raises(ConfigurationError):
        config.test

    config['sub.test'] = "1"
    assert config.sub.test == "1"
    config['sub.test'] = 1
    assert config.sub.test == "1"


def test_set_invalid_subconfig_value_like_dict():
    """Test that deep keys cannot be set if subconfig does not exist"""
    config = Configuration()
    with pytest.raises(BaseException) as exc:
        config['sub.test'] = "1"
    assert 'Configuration does not have an attribute \'sub\'.' in str(exc.value)


def test_yaml_loading_empty_config(yaml_path):
    """Test loading for empty config fails like setting non existing attributes."""
    config = Configuration()

    with pytest.raises(ConfigurationError) as exc:
        config.load_yaml(yaml_path)

    assert 'Configuration does not have an attribute \'test\'.' in str(exc.value)


def test_default_value():
    """Test that default value is given only when nothing else is available"""
    config = Configuration()
    config.add_option('test', option_type=str, default="voici_voila")
    assert config.test == "voici_voila"


def test_yaml_precedence(yaml_path):
    """Test that yaml definition has precedence over default values"""
    config = Configuration()
    config.add_option('test', option_type=str, default="voici_voila", env_var="TOP_SECRET_MESSAGE")
    assert config.test == "voici_voila"

    config.load_yaml(yaml_path)
    assert config.test == "from my yaml!"


def test_env_var_precedence(yaml_path):
    """Test that env_var has precedence over yaml values"""
    config = Configuration()
    config.add_option('test', option_type=str, default="voici_voila", env_var="TOP_SECRET_MESSAGE")
    assert config.test == "voici_voila"

    config.load_yaml(yaml_path)
    assert config.test == "from my yaml!"

    os.environ['TOP_SECRET_MESSAGE'] = 'coussi_coussa'
    assert config.test == "coussi_coussa"
    del os.environ['TOP_SECRET_MESSAGE']

    assert config.test == "from my yaml!"


def test_local_precedence(yaml_path):
    """Test local setting has precedence over env var values"""
    config = Configuration()
    config.add_option('test', option_type=str, default="voici_voila", env_var="TOP_SECRET_MESSAGE")
    assert config.test == "voici_voila"

    config.load_yaml(yaml_path)
    assert config.test == "from my yaml!"

    os.environ['TOP_SECRET_MESSAGE'] = 'coussi_coussa'
    assert config.test == "coussi_coussa"

    config.test = "comme_ci_comme_ca"
    assert config.test == "comme_ci_comme_ca"

    del os.environ['TOP_SECRET_MESSAGE']


def test_overwrite_subconfig():
    """Test that subconfig cannot be overwritten"""
    config = Configuration()
    config.nested = Configuration()
    with pytest.raises(ValueError) as exc:
        config.add_option('nested', option_type=str)
    assert "Configuration already contains nested" == str(exc.value)

    with pytest.raises(ValueError) as exc:
        config.nested = Configuration()
    assert "Configuration already contains subconfiguration nested" == str(exc.value)


def test_load_yaml_with_dict_items(subdict_yaml_path):
    """Test that yaml config with items assigned with dicts is supported"""
    config = Configuration()
    default = {'default': 'sub-dict'}
    config.add_option('test', option_type=dict, default=default)
    assert config.test == default
    config.test2 = Configuration()
    config.test2.add_option('test', option_type=dict, default=default)
    assert config.test2.test == default

    config.load_yaml(subdict_yaml_path)
    assert config.test == {'i am': 'a sub-dict'}
    assert config.test2.test == {'me is': 'sub-conf sub-dict'}


def test_load_yaml_unknown_option(broken_yaml_path):
    """Test error message when yaml config contains unknown options"""
    config = Configuration()
    config.add_option('test', option_type=str, default='hello')
    assert config.test == 'hello'

    with pytest.raises(ConfigurationError) as exc:
        config.load_yaml(broken_yaml_path)
    assert exc.match('Configuration does not have an attribute \'coucou\'')


def test_to_dict():
    """Test dictionary representation of the configuration"""
    config = Configuration()
    config.add_option('test', option_type=str, default="voici_voila")
    config.nested = Configuration()
    config.nested.add_option('test2', option_type=str, default="zici")

    assert config.to_dict() == {
        'test': 'voici_voila',
        'nested': {
            'test2': 'zici'}}

    config.test = 'hello'
    config.nested.test2 = 'labas'

    assert config.to_dict() == {
        'test': 'hello',
        'nested': {
            'test2': 'labas'}}


def test_key_curation():
    """Test that both - and _ maps to same options"""
    config = Configuration()
    config.add_option('test-with-dashes', option_type=int, default=1)
    config.add_option('test_with_underscores', option_type=int, default=2)
    config.add_option('test-all_mixedup', option_type=int, default=3)

    assert config['test-with-dashes'] == 1
    assert config['test_with_underscores'] == 2
    assert config['test-all_mixedup'] == 3

    assert config['test_with_dashes'] == 1
    assert config['test-with-underscores'] == 2
    assert config['test_all-mixedup'] == 3

    assert config.test_with_dashes == 1
    assert config.test_with_underscores == 2
    assert config.test_all_mixedup == 3

    config['test_with_dashes'] = 4
    assert config['test-with-dashes'] == 4


def test_nested_key_curation():
    """Test that both - and _ maps to same options in nested configs as well"""
    config = Configuration()
    config.add_option('test-with-dashes', option_type=str, default="voici_voila")
    config.nested = Configuration()
    config.nested.add_option('test_with_underscores', option_type=str, default="zici")

    assert config['nested']['test_with_underscores'] == 'zici'
    assert config['nested']['test-with-underscores'] == 'zici'

    config['nested']['test-with-underscores'] = 'labas'
    assert config.nested.test_with_underscores == 'labas'


def test_help_option():
    """Verify adding documentation to options."""
    config = Configuration()
    config.add_option('option', option_type=str, help='A useless option!')

    assert config.help('option') == 'A useless option!'


def test_help_nested_option():
    """Verify adding documentation to a nested option."""
    config = Configuration()
    config.add_option('option', option_type=str, help='A useless option!')
    config.nested = Configuration()
    config.nested.add_option('option', option_type=str, help='A useless nested option!')

    assert config.help('nested.option') == 'A useless nested option!'
    assert config.nested.help('option') == 'A useless nested option!'


def test_help_option_with_default():
    """Verify adding documentation to options with default value."""
    config = Configuration()
    config.add_option('option', option_type=str, default='a', help='A useless option!')

    assert config.help('option') == 'A useless option! (default: a)'


def test_no_help_option():
    """Verify not adding documentation to options."""
    config = Configuration()
    config.add_option('option', option_type=str)

    assert config.help('option') == 'Undocumented'


def test_argument_parser():
    """Verify the argument parser built based on config."""
    config = Configuration()
    config.add_option('option', option_type=str)

    parser = argparse.ArgumentParser()
    config.add_arguments(parser)

    options = parser.parse_args(['--option', 'a'])

    assert options.option == 'a'


def test_argument_parser_ignore_default():
    """Verify the argument parser does not get default values."""
    config = Configuration()
    config.add_option('option', option_type=str, default='b')

    parser = argparse.ArgumentParser()
    config.add_arguments(parser)

    options = parser.parse_args([])

    assert options.option is None


def test_argument_parser_rename():
    """Verify the argument parser built based on config with some options renamed."""
    config = Configuration()
    config.add_option('option', option_type=str)

    parser = argparse.ArgumentParser()
    config.add_arguments(parser, rename=dict(option='--new-option'))

    with pytest.raises(SystemExit) as exc:
        options = parser.parse_args(['--option', 'a'])

    assert exc.match('2')

    options = parser.parse_args(['--new-option', 'a'])

    assert options.new_option == 'a'


def test_argument_parser_dict_list_tuple():
    """Verify the argument parser does not contain options of type dict/list/tuple in config."""
    config = Configuration()
    config.add_option('st', option_type=str)
    config.add_option('di', option_type=dict)
    config.add_option('li', option_type=list)
    config.add_option('tu', option_type=tuple)

    parser = argparse.ArgumentParser()
    config.add_arguments(parser)

    options = parser.parse_args([])
    assert vars(options) == {'st': None}

    with pytest.raises(SystemExit) as exc:
        options = parser.parse_args(['--di', 'ct'])

    assert exc.match('2')


def test_deprecate_option(caplog):
    """Test deprecating an option."""
    config = Configuration()
    config.add_option(
        'option', option_type=str, default='hello',
        deprecate=dict(version='v1.0', alternative='None! T_T'))

    config.add_option(
        'ok', option_type=str, default='hillo')

    # Access the deprecated option and trigger a warning.
    with caplog.at_level(logging.WARNING, logger="orion.core.io.config"):
        assert config.option == 'hello'

    assert caplog.record_tuples == [(
        'orion.core.io.config', logging.WARNING,
        '(DEPRECATED) Option `option` will be removed in v1.0. Use `None! T_T` instead.')]

    caplog.clear()

    # Access the non-deprecated option and trigger no warnings.
    with caplog.at_level(logging.WARNING, logger="orion.core.io.config"):
        assert config.ok == 'hillo'

    assert caplog.record_tuples == []


def test_deprecate_option_missing_version():
    """Verify option deprecation if version is missing."""
    config = Configuration()
    with pytest.raises(ValueError) as exc:
        config.add_option(
            'option', option_type=str,
            deprecate=dict(alternative='None! T_T'))

    assert exc.match('`version` is missing in deprecate option')


def test_deprecate_option_no_alternative(caplog):
    """Verify option deprecation when there is no alternative."""
    config = Configuration()
    config.add_option(
        'option', option_type=str, default='hello',
        deprecate=dict(version='v1.0'))

    # Access the deprecated option and trigger a warning.
    with caplog.at_level(logging.WARNING, logger="orion.core.io.config"):
        assert config.option == 'hello'

    assert caplog.record_tuples == [(
        'orion.core.io.config', logging.WARNING,
        '(DEPRECATED) Option `option` will be removed in v1.0.')]


def test_deprecate_option_help():
    """Verify help message of a deprecated option."""
    config = Configuration()
    config.add_option(
        'option', option_type=str,
        deprecate=dict(version='v1.0', alternative='None! T_T'),
        help='A useless option!')

    assert config.help('option') == '(DEPRECATED) A useless option!'


def test_deprecate_option_print_with_different_name(caplog):
    """Verify deprecation warning with different name (for nested options)."""
    config = Configuration()
    config.add_option(
        'option', option_type=str, default='hello',
        deprecate=dict(version='v1.0', alternative='None! T_T', name='nested.option'))

    # Access the deprecated option and trigger a warning.
    with caplog.at_level(logging.WARNING, logger="orion.core.io.config"):
        assert config.option == 'hello'

    assert caplog.record_tuples == [(
        'orion.core.io.config', logging.WARNING,
        '(DEPRECATED) Option `nested.option` will be removed in v1.0. Use `None! T_T` instead.')]


def test_get_deprecated_key(caplog):
    """Verify deprecation warning using get()."""
    config = Configuration()
    config.add_option(
        'option', option_type=str, default='hello',
        deprecate=dict(version='v1.0', alternative='None! T_T'))

    # Access the deprecated option and trigger a warning.
    with caplog.at_level(logging.WARNING, logger="orion.core.io.config"):
        assert config.get('option') == 'hello'

    assert caplog.record_tuples == [(
        'orion.core.io.config', logging.WARNING,
        '(DEPRECATED) Option `option` will be removed in v1.0. Use `None! T_T` instead.')]


def test_get_deprecated_key_ignore_warning(caplog):
    """Verify deprecation warning using get(deprecated='ignore')."""
    config = Configuration()
    config.add_option(
        'option', option_type=str, default='hello',
        deprecate=dict(version='v1.0', alternative='None! T_T'))

    # Access the deprecated option and trigger a warning.
    with caplog.at_level(logging.WARNING, logger="orion.core.io.config"):
        assert config.get('option', deprecated='ignore') == 'hello'

    assert caplog.record_tuples == []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests resolutions in :mod:`orion.core.evc.conflicts`."""

import pytest

from orion.algo.space import Dimension
from orion.core.evc import adapters
from orion.core.evc import conflicts


@pytest.fixture
def add_dimension_resolution(new_dimension_conflict):
    """Create a resolution for a new dimension"""
    return new_dimension_conflict.AddDimensionResolution(new_dimension_conflict)


@pytest.fixture
def change_dimension_resolution(changed_dimension_conflict):
    """Create a resolution for a changed prior"""
    return changed_dimension_conflict.ChangeDimensionResolution(changed_dimension_conflict)


@pytest.fixture
def remove_dimension_resolution(missing_dimension_conflict):
    """Create a resolution to remove a missing dimension"""
    return missing_dimension_conflict.RemoveDimensionResolution(missing_dimension_conflict,
                                                                default_value=0)


@pytest.fixture
def rename_dimension_resolution(missing_dimension_conflict, new_dimension_conflict):
    """Create a resolution to rename a missing dimension to a new dimension"""
    return missing_dimension_conflict.RenameDimensionResolution(missing_dimension_conflict,
                                                                new_dimension_conflict)


@pytest.fixture
def algorithm_resolution(algorithm_conflict):
    """Create a resolution for a new algorithm configuration"""
    return algorithm_conflict.AlgorithmResolution(algorithm_conflict)


@pytest.fixture
def code_resolution(code_conflict):
    """Create a resolution for a code conflict"""
    return code_conflict.CodeResolution(code_conflict, adapters.CodeChange.BREAK)


@pytest.fixture
def experiment_name_resolution(create_db_instance, experiment_name_conflict):
    """Create a resolution for a code conflict"""
    return experiment_name_conflict.ExperimentNameResolution(experiment_name_conflict,
                                                             new_name="new-exp-name")


class TestAddDimensionResolution(object):
    """Test methods for resolution of new dimension"""

    def test_init_no_default(self, new_dimension_conflict):
        """Verify instantiation of resolution without default value"""
        resolution = new_dimension_conflict.AddDimensionResolution(new_dimension_conflict)
        assert resolution.default_value is Dimension.NO_DEFAULT_VALUE

    def test_init_no_default_but_in_dim(self, new_dimension_with_default_conflict):
        """Verify instantiation of resolution with default value in dimension"""
        resolution = new_dimension_with_default_conflict.AddDimensionResolution(
            new_dimension_with_default_conflict)
        assert (resolution.default_value ==
                new_dimension_with_default_conflict.dimension.default_value)

    def test_init_with_default(self, new_dimension_conflict,
                               new_dimension_with_default_conflict):
        """Verify instantiation of resolution with default value given by user"""
        default_value = 1.1
        resolution = new_dimension_conflict.AddDimensionResolution(
            new_dimension_conflict, default_value=default_value)
        assert resolution.default_value == default_value

        assert new_dimension_with_default_conflict.dimension.default_value != default_value
        resolution = new_dimension_conflict.AddDimensionResolution(
            new_dimension_with_default_conflict, default_value=default_value)
        assert resolution.default_value == default_value

    def test_init_bad_default(self, new_dimension_conflict):
        """Verify instantiation of resolution with invalid default value"""
        default_value = 'bad'
        with pytest.raises(ValueError) as exc:
            new_dimension_conflict.AddDimensionResolution(new_dimension_conflict,
                                                          default_value=default_value)
        assert "could not convert string to float: 'bad'" in str(exc.value)

    def test_new_prior_no_default(self, new_dimension_conflict):
        """Verify prior string without default value"""
        resolution = new_dimension_conflict.AddDimensionResolution(new_dimension_conflict)
        assert resolution.new_prior == 'norm(0, 2)'

    def test_new_prior_default_from_dim(self, new_dimension_with_default_conflict):
        """Verify prior string with default value in dimension"""
        resolution = new_dimension_with_default_conflict.AddDimensionResolution(
            new_dimension_with_default_conflict)
        assert resolution.new_prior == 'norm(0, 2, default_value=0.001)'

    def test_new_prior_default(self, new_dimension_conflict,
                               new_dimension_with_default_conflict):
        """Verify prior string with new default value given by user"""
        default_value = 1.2
        resolution = new_dimension_with_default_conflict.AddDimensionResolution(
            new_dimension_with_default_conflict, default_value=default_value)
        assert resolution.new_prior == 'norm(0, 2, default_value={})'.format(default_value)

        resolution = new_dimension_conflict.AddDimensionResolution(new_dimension_conflict,
                                                                   default_value=default_value)
        assert resolution.new_prior == 'norm(0, 2, default_value={})'.format(default_value)

    def test_prefix(self, add_dimension_resolution):
        """Verify prefix of resolution with corresponding marker"""
        assert add_dimension_resolution.prefix == "new~+"

    def test_repr_without_default(self, add_dimension_resolution):
        """Verify resolution representation for user interface, without default value"""
        assert repr(add_dimension_resolution) == "new~+norm(0, 2)"

    def test_repr_default_from_dim(self, new_dimension_with_default_conflict):
        """Verify resolution representation for user interface, without default value"""
        resolution = new_dimension_with_default_conflict.AddDimensionResolution(
            new_dimension_with_default_conflict)
        assert repr(resolution) == "new~+norm(0, 2, default_value=0.001)"

    def test_adapters_without_default(self, new_dimension_conflict):
        """Verify adapters without default values (filter everything out)"""
        param = {'name': 'new', 'type': 'real', 'value': Dimension.NO_DEFAULT_VALUE}
        resolution = new_dimension_conflict.AddDimensionResolution(new_dimension_conflict)
        resolution_adapters = resolution.get_adapters()
        assert len(resolution_adapters) == 1
        assert (resolution_adapters[0].configuration ==
                adapters.DimensionAddition(param).configuration)

    def test_adapters_with_default(self, new_dimension_conflict):
        """Verify adapters with default values"""
        param = {'name': 'new', 'type': 'real', 'value': 1.1}
        resolution = new_dimension_conflict.AddDimensionResolution(new_dimension_conflict,
                                                                   default_value=1.1)
        resolution_adapters = resolution.get_adapters()
        assert len(resolution_adapters) == 1
        assert (resolution_adapters[0].configuration ==
                adapters.DimensionAddition(param).configuration)

    def test_revert(self, new_dimension_conflict, add_dimension_resolution):
        """Verify reverting resolution set conflict to unresolved"""
        assert new_dimension_conflict.is_resolved
        assert add_dimension_resolution.revert() == []
        assert not new_dimension_conflict.is_resolved
        assert new_dimension_conflict.resolution is None


class TestChangeDimensionResolution(object):
    """Test methods for resolution of changed dimensions"""

    def test_prefix(self, change_dimension_resolution):
        """Verify prefix of resolution with corresponding marker"""
        assert change_dimension_resolution.prefix == 'changed~+'

    def test_repr(self, change_dimension_resolution):
        """Verify resolution representation for user interface"""
        assert repr(change_dimension_resolution) == 'changed~+normal(0, 2)'

    def test_adapters(self, change_dimension_resolution):
        """Verify adapters with old and new priors"""
        name = "changed"
        old_prior = "uniform(-10, 10)"
        new_prior = "normal(0, 2)"
        resolution_adapters = change_dimension_resolution.get_adapters()
        assert len(resolution_adapters) == 1
        assert (resolution_adapters[0].configuration ==
                adapters.DimensionPriorChange(name, old_prior, new_prior).configuration)

    def test_revert(self, changed_dimension_conflict, change_dimension_resolution):
        """Verify reverting resolution set conflict to unresolved"""
        assert changed_dimension_conflict.is_resolved
        assert change_dimension_resolution.revert() == []
        assert not changed_dimension_conflict.is_resolved
        assert changed_dimension_conflict.resolution is None


class TestRemoveDimensionResolution(object):
    """Test methods for resolution of missing dimensions"""

    def test_prefix(self, missing_dimension_conflict):
        """Verify prefix of resolution with corresponding marker"""
        resolution = missing_dimension_conflict.RemoveDimensionResolution(
            missing_dimension_conflict)
        assert resolution.prefix == 'missing~-'

    def test_repr_no_default(self, missing_dimension_conflict):
        """Verify resolution representation for user interface, without default value"""
        resolution = missing_dimension_conflict.RemoveDimensionResolution(
            missing_dimension_conflict)
        assert repr(resolution) == 'missing~-'

    def test_repr_default_from_dim(self, missing_dimension_with_default_conflict):
        """Verify resolution representation for user interface, with default value from dimension"""
        resolution = missing_dimension_with_default_conflict.RemoveDimensionResolution(
            missing_dimension_with_default_conflict)
        assert repr(resolution) == 'missing~-0.0'

    def test_repr_default(self, missing_dimension_conflict,
                          missing_dimension_with_default_conflict):
        """Verify resolution representation for user interface, with default provided by user"""
        default_value = 1.2
        resolution = missing_dimension_conflict.RemoveDimensionResolution(
            missing_dimension_with_default_conflict, default_value=default_value)
        assert repr(resolution) == 'missing~-{}'.format(default_value)

        resolution = missing_dimension_conflict.RemoveDimensionResolution(
            missing_dimension_conflict, default_value=default_value)
        assert repr(resolution) == 'missing~-{}'.format(default_value)

    def test_adapters_without_default(self, missing_dimension_conflict):
        """Verify adapters without default value"""
        param = {'name': 'missing', 'type': 'real', 'value': Dimension.NO_DEFAULT_VALUE}
        resolution = missing_dimension_conflict.RemoveDimensionResolution(
            missing_dimension_conflict)
        resolution_adapters = resolution.get_adapters()
        assert len(resolution_adapters) == 1
        assert (resolution_adapters[0].configuration ==
                adapters.DimensionDeletion(param).configuration)

    def test_adapters_with_default(self, missing_dimension_conflict):
        """Verify adapters with default value"""
        param = {'name': 'missing', 'type': 'real', 'value': 1.2}
        resolution = missing_dimension_conflict.RemoveDimensionResolution(
            missing_dimension_conflict, default_value=1.2)
        resolution_adapters = resolution.get_adapters()
        assert len(resolution_adapters) == 1
        assert (resolution_adapters[0].configuration ==
                adapters.DimensionDeletion(param).configuration)

    def test_revert(self, missing_dimension_conflict, remove_dimension_resolution):
        """Verify reverting resolution set conflict to unresolved"""
        assert missing_dimension_conflict.is_resolved
        assert remove_dimension_resolution.revert() == []
        assert not missing_dimension_conflict.is_resolved
        assert missing_dimension_conflict.resolution is None


class TestRenameDimensionResolution(object):
    """Test methods for renaming of missing dimensions"""

    def test_init_same_prior(self, missing_dimension_conflict, new_dimension_same_prior_conflict):
        """Verify initialisation with identical priors generates no side-effect conflicts"""
        assert not missing_dimension_conflict.is_resolved
        assert not new_dimension_same_prior_conflict.is_resolved
        resolution = missing_dimension_conflict.RenameDimensionResolution(
            missing_dimension_conflict, new_dimension_same_prior_conflict)
        assert missing_dimension_conflict.is_resolved
        assert new_dimension_same_prior_conflict.is_resolved
        assert resolution.new_conflicts == []

    def test_init_different_prior(self, missing_dimension_conflict, new_dimension_conflict):
        """Verify initialisation with different priors generates a side-effect conflict"""
        assert not missing_dimension_conflict.is_resolved
        assert not new_dimension_conflict.is_resolved
        resolution = missing_dimension_conflict.RenameDimensionResolution(
            missing_dimension_conflict, new_dimension_conflict)
        assert missing_dimension_conflict.is_resolved
        assert new_dimension_conflict.is_resolved
        assert len(resolution.new_conflicts) == 1
        assert isinstance(resolution.new_conflicts[0], conflicts.ChangedDimensionConflict)
        assert resolution.new_conflicts[0].old_prior == missing_dimension_conflict.prior
        assert resolution.new_conflicts[0].new_prior == new_dimension_conflict.prior

    def test_prefix(self, rename_dimension_resolution):
        """Verify prefix of resolution with corresponding marker"""
        assert rename_dimension_resolution.prefix == 'missing~>'

    def test_repr(self, rename_dimension_resolution):
        """Verify resolution representation for user interface"""
        assert repr(rename_dimension_resolution) == 'missing~>new'

    def test_adapters(self, rename_dimension_resolution):
        """Verify adapters with old and new names"""
        old_name = "missing"
        new_name = "new"
        resolution_adapters = rename_dimension_resolution.get_adapters()
        assert len(resolution_adapters) == 1
        assert (resolution_adapters[0].configuration ==
                adapters.DimensionRenaming(old_name, new_name).configuration)

    def test_revert_same_prior(self, missing_dimension_conflict, new_dimension_same_prior_conflict):
        """Verify reverting resolution set conflict to unresolved"""
        resolution = missing_dimension_conflict.RenameDimensionResolution(
            missing_dimension_conflict, new_dimension_same_prior_conflict)
        assert missing_dimension_conflict.is_resolved
        assert new_dimension_same_prior_conflict.is_resolved
        assert len(resolution.new_conflicts) == 0
        assert resolution.revert() == []
        assert len(resolution.new_conflicts) == 0
        assert not missing_dimension_conflict.is_resolved
        assert not new_dimension_same_prior_conflict.is_resolved
        assert missing_dimension_conflict.resolution is None
        assert new_dimension_same_prior_conflict.resolution is None

    def test_revert_different_prior(self, missing_dimension_conflict, new_dimension_conflict,
                                    rename_dimension_resolution):
        """Verify reverting resolution set conflict to unresolved and deprecate the side-effect
        conflict
        """
        assert missing_dimension_conflict.is_resolved
        assert new_dimension_conflict.is_resolved
        assert len(rename_dimension_resolution.new_conflicts) == 1
        new_conflicts = rename_dimension_resolution.new_conflicts
        assert rename_dimension_resolution.revert() == new_conflicts
        assert len(rename_dimension_resolution.new_conflicts) == 0
        assert new_conflicts[0].is_resolved
        assert not missing_dimension_conflict.is_resolved
        assert not new_dimension_conflict.is_resolved
        assert missing_dimension_conflict.resolution is None
        assert new_dimension_conflict.resolution is None


class TestAlgorithmResolution(object):
    """Test methods for resolution of algorithm changes"""

    def test_adapters(self, algorithm_resolution):
        """Verify shallow adapters for algorithm change"""
        resolution_adapters = algorithm_resolution.get_adapters()
        assert len(resolution_adapters) == 1
        assert resolution_adapters[0].configuration == adapters.AlgorithmChange().configuration

    def test_repr(self, algorithm_resolution):
        """Verify resolution representation for user interface"""
        assert repr(algorithm_resolution) == '--algorithm-change'

    def test_revert(self, algorithm_conflict, algorithm_resolution):
        """Verify reverting resolution set conflict to unresolved"""
        assert algorithm_conflict.is_resolved
        assert algorithm_resolution.revert() == []
        assert not algorithm_conflict.is_resolved
        assert algorithm_conflict.resolution is None


class TestCodeResolution(object):
    """Test methods for resolution of code conflict"""

    def test_wrong_input(self, code_conflict):
        """Verify initialization fails with invalid change type"""
        with pytest.raises(ValueError) as exc:
            code_conflict.CodeResolution(code_conflict, "yabadabadoo")
        assert "Invalid code change type" in str(exc.value)

    def test_adapters(self, code_conflict):
        """Verify adapters with code change types"""
        for change_type in adapters.CodeChange.types:
            code_resolution = code_conflict.CodeResolution(code_conflict, change_type)
            resolution_adapters = code_resolution.get_adapters()
            assert len(resolution_adapters) == 1
            assert (resolution_adapters[0].configuration ==
                    adapters.CodeChange(change_type).configuration)

    def test_repr(self, code_resolution):
        """Verify resolution representation for user interface"""
        assert repr(code_resolution) == '--code-change-type break'

    def test_revert(self, code_conflict, code_resolution):
        """Verify reverting resolution set conflict to unresolved"""
        assert code_conflict.is_resolved
        assert code_resolution.revert() == []
        assert not code_conflict.is_resolved
        assert code_conflict.resolution is None


class TestExperimentNameResolution(object):
    """Test methods for resolution of experiment name conflict"""

    def test_adapters(self, experiment_name_resolution):
        """Verify there is no adapters for experiment name resolution"""
        assert experiment_name_resolution.get_adapters() == []

    def test_repr(self, experiment_name_resolution):
        """Verify resolution representation for user interface"""
        assert repr(experiment_name_resolution) == '--branch-to new-exp-name'

    def test_revert(self, old_config, new_config,
                    experiment_name_conflict, experiment_name_resolution):
        """Verify reverting resolution set conflict to unresolved and reset name in config"""
        assert experiment_name_conflict.is_resolved
        assert new_config['name'] == experiment_name_resolution.new_name
        assert new_config['version'] == experiment_name_resolution.new_version

        assert experiment_name_resolution.revert() == []

        assert new_config['name'] == old_config['name']
        assert new_config['version'] == old_config['version']
        assert not experiment_name_conflict.is_resolved
        assert experiment_name_conflict.resolution is None



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.worker.primary_algo` -- Sanitizing wrapper of main algorithm
=============================================================================

.. module:: primary_algo
   :platform: Unix
   :synopsis: Performs checks and organizes required transformations of points.

"""
from orion.algo.base import BaseAlgorithm
from orion.core.worker.transformer import build_required_space


# pylint: disable=too-many-public-methods
class PrimaryAlgo(BaseAlgorithm):
    """Perform checks on points and transformations. Wrap the primary algorithm.

    1. Checks requirements on the parameter space from algorithms and create the
    appropriate transformations. Apply transformations before and after methods
    of the primary algorithm.
    2. Checks whether incoming and outcoming points are compliant with a space.

    """

    def __init__(self, space, algorithm_config):
        """
        Initialize the primary algorithm.

        Parameters
        ----------
        space : `orion.algo.space.Space`
           The original definition of a problem's parameters space.
        algorithm_config : dict
           Configuration for the algorithm.

        """
        self.algorithm = None
        super(PrimaryAlgo, self).__init__(space, algorithm=algorithm_config)
        requirements = self.algorithm.requires
        self.transformed_space = build_required_space(requirements, self.space)
        self.algorithm.space = self.transformed_space

    def seed_rng(self, seed):
        """Seed the state of the algorithm's random number generator."""
        self.algorithm.seed_rng(seed)

    @property
    def state_dict(self):
        """Return a state dict that can be used to reset the state of the algorithm."""
        return self.algorithm.state_dict

    def set_state(self, state_dict):
        """Reset the state of the algorithm based on the given state_dict

        :param state_dict: Dictionary representing state of an algorithm
        """
        self.algorithm.set_state(state_dict)

    def suggest(self, num=1):
        """Suggest a `num` of new sets of parameters.

        :param num: how many sets to be suggested.

        .. note:: New parameters must be compliant with the problem's domain
           `orion.algo.space.Space`.
        """
        points = self.algorithm.suggest(num)

        if points is None:
            return None

        for point in points:
            if point not in self.transformed_space:
                raise ValueError("""
Point is not contained in space:
Point: {}
Space: {}""".format(point, self.transformed_space))

        return [self.transformed_space.reverse(point) for point in points]

    def observe(self, points, results):
        """Observe evaluation `results` corresponding to list of `points` in
        space.

        .. seealso:: `orion.algo.base.BaseAlgorithm.observe`
        """
        assert len(points) == len(results)
        tpoints = []
        for point in points:
            assert point in self.space
            tpoints.append(self.transformed_space.transform(point))
        self.algorithm.observe(tpoints, results)

    @property
    def is_done(self):
        """Return True, if an algorithm holds that there can be no further improvement."""
        return self.algorithm.is_done

    def score(self, point):
        """Allow algorithm to evaluate `point` based on a prediction about
        this parameter set's performance. Return a subjective measure of expected
        performance.

        By default, return the same score any parameter (no preference).
        """
        assert point in self.space
        return self.algorithm.score(self.transformed_space.transform(point))

    def judge(self, point, measurements):
        """Inform an algorithm about online `measurements` of a running trial.

        The algorithm can return a dictionary of data which will be provided
        as a response to the running environment. Default is None response.

        """
        assert point in self._space
        return self.algorithm.judge(self.transformed_space.transform(point),
                                    measurements)

    @property
    def should_suspend(self):
        """Allow algorithm to decide whether a particular running trial is still
        worth to complete its evaluation, based on information provided by the
        `judge` method.

        """
        return self.algorithm.should_suspend

    @property
    def configuration(self):
        """Return tunable elements of this algorithm in a dictionary form
        appropriate for saving.
        """
        return self.algorithm.configuration

    @property
    def space(self):
        """Domain of problem associated with this algorithm's instance.

        .. note:: Redefining property here without setter, denies base class' setter.
        """
        return self._space



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Test base functionalities of :mod:`orion.core.utils`."""

import pytest

from orion.core.utils import Factory


def test_factory_subclasses_detection():
    """Verify that meta-class Factory finds all subclasses"""
    class Base(object):
        pass

    class A(Base):
        pass

    class B(Base):
        pass

    class AA(A):
        pass

    class AB(A):
        pass

    class AAA(AA):
        pass

    class AA_AB(AA, AB):
        pass

    class MyFactory(Base, metaclass=Factory):
        pass

    assert type(MyFactory(of_type="A")) is A
    assert type(MyFactory(of_type="B")) is B
    assert type(MyFactory(of_type="AA")) is AA
    assert type(MyFactory(of_type="AAA")) is AAA
    assert type(MyFactory(of_type="AA_AB")) is AA_AB

    # Test if there is duplicates
    assert set(MyFactory.types) == set((A, B, AA, AB, AAA, AA_AB))
    assert len(MyFactory.types) == len(set(MyFactory.types))

    with pytest.raises(NotImplementedError) as exc_info:
        MyFactory(of_type="random")
    assert "Could not find implementation of Base, type = 'random'" in str(exc_info.value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Example usage and tests for :mod:`orion.core.worker.primary_algo`."""

import pytest

from orion.core.worker.primary_algo import PrimaryAlgo


@pytest.fixture()
def palgo(dumbalgo, space, fixed_suggestion):
    """Set up a PrimaryAlgo with dumb configuration."""
    algo_config = {'DumbAlgo': {
        'value': fixed_suggestion,
        'subone': {'DumbAlgo': dict(
            value=6,
            scoring=5
            )}
        }}
    palgo = PrimaryAlgo(space, algo_config)

    return palgo


class TestPrimaryAlgoWraps(object):
    """Test if PrimaryAlgo is actually wrapping the configured algorithm.

    Does not test for transformations.
    """

    def test_init_and_configuration(self, dumbalgo, palgo, fixed_suggestion):
        """Check if initialization works."""
        assert isinstance(palgo.algorithm, dumbalgo)
        assert palgo.configuration == {
            'dumbalgo': {
                'seed': None,
                'value': fixed_suggestion,
                'scoring': 0,
                'judgement': None,
                'suspend': False,
                'done': False,
                'subone': {
                    'dumbalgo': {
                        'seed': None,
                        'value': 6,
                        'scoring': 5,
                        'judgement': None,
                        'suspend': False,
                        'done': False,
                        }
                    }
                }
            }

    def test_space_can_only_retrieved(self, palgo, space):
        """Set space is forbidden, getter works as supposed."""
        assert palgo.space == space
        with pytest.raises(AttributeError):
            palgo.space = 5

    def test_suggest(self, palgo, fixed_suggestion):
        """Suggest wraps suggested."""
        assert palgo.suggest() == [fixed_suggestion]
        assert palgo.suggest(4) == [fixed_suggestion] * 4
        palgo.algorithm.possible_values = [(5,)]
        with pytest.raises(ValueError):
            palgo.suggest()

    def test_observe(self, palgo, fixed_suggestion):
        """Observe wraps observations."""
        palgo.observe([fixed_suggestion], [5])
        assert palgo.algorithm._points == [fixed_suggestion]
        assert palgo.algorithm._results == [5]
        with pytest.raises(AssertionError):
            palgo.observe([fixed_suggestion], [5, 8])
        with pytest.raises(AssertionError):
            palgo.observe([(5,)], [5])

    def test_isdone(self, palgo):
        """Wrap isdone."""
        palgo.algorithm.done = 10
        assert palgo.is_done == 10
        assert palgo.algorithm._times_called_is_done == 1

    def test_shouldsuspend(self, palgo):
        """Wrap should_suspend."""
        palgo.algorithm.suspend = 55
        assert palgo.should_suspend == 55
        assert palgo.algorithm._times_called_suspend == 1

    def test_score(self, palgo, fixed_suggestion):
        """Wrap score."""
        palgo.algorithm.scoring = 60
        assert palgo.score(fixed_suggestion) == 60
        assert palgo.algorithm._score_point == fixed_suggestion
        with pytest.raises(AssertionError):
            palgo.score((5,))

    def test_judge(self, palgo, fixed_suggestion):
        """Wrap judge."""
        palgo.algorithm.judgement = 'naedw'
        assert palgo.judge(fixed_suggestion, 8) == 'naedw'
        assert palgo.algorithm._judge_point == fixed_suggestion
        assert palgo.algorithm._measurements == 8
        with pytest.raises(AssertionError):
            palgo.judge((5,), 8)


class TestPrimaryAlgoTransforms(object):
    """Check if PrimaryAlgo appropriately transforms spaces and samples."""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.serving.webapi` -- WSGI REST server application
===========================================================

.. module:: webapi
   :platform: Unix
   :synopsis: Exposes a WSGI REST server application instance by subclassing ``falcon.API``
"""

import falcon

from orion.serving.experiments_resource import ExperimentsResource
from orion.serving.plots_resources import PlotsResource
from orion.serving.runtime import RuntimeResource
from orion.serving.trials_resource import TrialsResource
from orion.storage.base import setup_storage


class WebApi(falcon.API):
    """
    Main entry point into a Falcon-based app. An instance provides a callable WSGI interface and a
    routing engine.
    """

    def __init__(self, config=None):
        super(WebApi, self).__init__()
        self.config = config

        setup_storage(config.get('storage'))

        # Create our resources
        root_resource = RuntimeResource()
        experiments_resource = ExperimentsResource()
        trials_resource = TrialsResource()
        plots_resource = PlotsResource()

        # Build routes
        self.add_route('/', root_resource)
        self.add_route('/experiments', experiments_resource)
        self.add_route('/experiments/{name}', experiments_resource, suffix="experiment")
        self.add_route('/trials/{experiment_name}', trials_resource, suffix='trials_in_experiment')
        self.add_route('/trials/{experiment_name}/{trial_id}',
                       trials_resource, suffix='trial_in_experiment')
        self.add_route('/plots/regret/{experiment_name}', plots_resource, suffix='regret')

    def start(self):
        """A hook to when a Gunicorn worker calls run()."""
        pass

    def stop(self, signal):
        """A hook to when a Gunicorn worker starts shutting down."""
        pass



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.worker.convert` -- Parse and generate user script's configuration
==================================================================================

.. module:: convert
   :platform: Unix
   :synopsis: Defines and instantiates a converter for configuration file types.

Given a file path infer which configuration file parser/emitter it corresponds to.
Define `Converter` classes with a common interface for many popular configuration
file types.

Currently supported:
    - YAML
    - JSON
    - See below, for configuration agnostic parsing

A `GenericConverter` is provided that tries and parses configuration
files, regardless of their type, according to predefined Oríon's markers.

"""
from abc import (ABC, abstractmethod)
from collections import deque
import importlib
import os

from orion.core.utils import (Factory, nesteddict)


def infer_converter_from_file_type(config_path,
                                   regex=None, default_keyword=''):
    """Use filetype extension to infer and build the correct configuration file
    converter.
    """
    _, ext_type = os.path.splitext(os.path.abspath(config_path))
    for klass in Converter.types:
        if ext_type in klass.file_extensions:
            return klass()

    if regex is None:
        return GenericConverter(expression_prefix=default_keyword)

    return GenericConverter(regex, expression_prefix=default_keyword)


class BaseConverter(ABC):
    """Base class for configuration parsers/generators.

    Attributes
    ----------
    file_extensions : list of strings
       Strings starting with '.' which identify usually a file type as a
       common convention. For instance, ``['.yml', '.yaml']`` for YAML files.

    """

    file_extensions = []

    # pylint:disable=no-self-use
    def get_state_dict(self):
        """Give state dict that can be used to reconstruct the converter"""
        return {}

    def set_state_dict(self, state):
        """Reset the converter based on previous state"""
        pass

    @abstractmethod
    def parse(self, filepath):
        """Read dictionary out of the configuration file.

        Parameters
        ----------
        filepath : str
           Full path to the original user script's configuration.

        """
        pass

    @abstractmethod
    def generate(self, filepath, data):
        """Create a configuration file at `filepath` using dictionary `data`."""
        pass


class YAMLConverter(BaseConverter):
    """Converter for YAML files."""

    file_extensions = ['.yml', '.yaml']

    def __init__(self):
        """Try to dynamically import yaml module."""
        self.yaml = importlib.import_module('yaml')

    def parse(self, filepath):
        """Read dictionary out of the configuration file.

        Parameters
        ----------
        file : str
           Full path to the original user script's configuration.

        """
        with open(filepath) as f:
            return self.yaml.safe_load(stream=f)

    def generate(self, filepath, data):
        """Create a configuration file at `filepath` using dictionary `data`."""
        with open(filepath, 'w') as f:
            self.yaml.dump(data, stream=f)


class JSONConverter(BaseConverter):
    """Converter for JSON files."""

    file_extensions = ['.json']

    def __init__(self):
        """Try to dynamically import json module."""
        self.json = importlib.import_module('json')

    def parse(self, filepath):
        """Read dictionary out of the configuration file.

        Parameters
        ----------
        file : str
           Full path to the original user script's configuration.

        """
        with open(filepath) as f:
            return self.json.load(f)

    def generate(self, filepath, data):
        """Create a configuration file at `filepath` using dictionary `data`."""
        with open(filepath, 'w') as f:
            self.json.dump(data, f)


class GenericConverter(BaseConverter):
    """Generic converter for any configuration file type.

    For each parameter dimension declared here, one must necessarily
    provide a ``name`` keyword inside the `Dimension` building expression.

    Implementation details: As this class is supposed to provide with a
    generic text parser, semantics are going to be tied to their consequent
    usage. A template document is going to be created on `parse` and filled
    with values on `read`. This template document consists the state of this
    `Converter` object.

    Dimension should be defined for instance as:
    ``meaningful_name~uniform(0, 4)``

    """

    def __init__(self, regex=r'([\/]?[\w|\/|-]+)~([\+]?.*\)|\-|\>[A-Za-z_]\w*)',
                 expression_prefix=''):
        """Initialize with the regex expression which will be searched for
        to define a `Dimension`.
        """
        self.re_module = importlib.import_module('re')
        self.regex = self.re_module.compile(regex)
        self.expression_prefix = expression_prefix
        self.template = None
        self.has_leading = dict()
        self.conflict_msg = "Namespace conflict in configuration file '{}', under '{}'"

    def get_state_dict(self):
        """Give state dict that can be used to reconstruct the converter"""
        return dict(
            regex=self.regex.pattern,
            expression_prefix=self.expression_prefix,
            template=self.template,
            has_leading=self.has_leading)

    def set_state_dict(self, state):
        """Reset the converter based on previous state"""
        self.regex = self.re_module.compile(state['regex'])
        self.expression_prefix = state['expression_prefix']
        self.template = state['template']
        self.has_leading = state['has_leading']

    def _raise_conflict(self, path, namespace):
        raise ValueError(self.conflict_msg.format(path, namespace))

    def parse(self, filepath):
        r"""Read dictionary out of the configuration file.

        Create a template for Python 3 string format and save it as this
        object's state, by substituing '{\1}' wherever the pattern
        was matched. By default, the first matched group (\1) corresponds
        with a dimension's namespace.

        .. note:: Namespace in substitution templates does not contain the first '/'.

        Parameters
        ----------
        filepath : str
           Full path to the original user script's configuration.

        """
        with open(filepath) as f:
            self.template = f.read()

        # Search for Oríon semantic pattern
        pairs = self.regex.findall(self.template)
        ret = dict(pairs)

        # Every namespace given should be unique,
        # raise conflict if there are duplicates
        if len(pairs) != len(ret):
            namespaces = list(zip(*pairs))[0]
            for name in namespaces:
                if namespaces.count(name) != 1:
                    self._raise_conflict(filepath, name)

        # Create template using each namespace as format key,
        # exactly as provided by the user
        subst = self.re_module.sub(r'{', r'{{', self.template)
        subst = self.re_module.sub(r'}', r'}}', subst)
        substituted, num_subs = self.regex.subn(r'{\1!s}', subst)
        assert len(ret) == num_subs, "This means an error in the regex. Report bug. Details::\n"\
            "original: {}\n, regex:{}".format(self.template, self.regex)
        self.template = substituted

        # Wrap it in style of what the rest of `Converter`s return
        ret_nested = nesteddict()
        for namespace, expression in ret.items():
            keys = namespace.split('/')
            if not keys[0]:  # It means that user wrote a namespace starting from '/'
                keys = keys[1:]  # Safe because of the regex pattern
                self.has_leading[namespace[1:]] = '/'

            stuff = ret_nested
            for i, key in enumerate(keys[:-1]):
                stuff = stuff[key]
                if isinstance(stuff, str):
                    # If `stuff` is not a dictionary while traversing the
                    # namespace path, then this amounts to a conflict which was
                    # not sufficiently get caught
                    self._raise_conflict(filepath, '/'.join(keys[:i + 1]))
            # If final value is already filled,
            # then this must be also due to a conflict
            if stuff[keys[-1]]:
                self._raise_conflict(filepath, namespace)

            # Keep compatibility with `SpaceBuilder._build_from_config`
            stuff[keys[-1]] = self.expression_prefix + expression

        return ret_nested

    def generate(self, filepath, data):
        """Create a configuration file at `filepath` using dictionary `data`."""
        unnested_data = dict()
        stack = deque()
        stack.append(([], data))
        while True:
            try:
                namespace, stuff = stack.pop()
            except IndexError:
                break
            if isinstance(stuff, dict):
                for k, v in stuff.items():
                    stack.append((['/'.join(namespace + [str(k)])], v))
            else:
                name = namespace[0]
                unnested_data[self.has_leading.get(name, '') + name] = stuff

        print(self.template)
        print(unnested_data)
        document = self.template.format(**unnested_data)

        with open(filepath, 'w') as f:
            f.write(document)


# pylint: disable=too-few-public-methods,abstract-method
class Converter(BaseConverter, metaclass=Factory):
    """Class used to inject dependency on a configuration file parser/generator.

    .. seealso:: `Factory` metaclass and `BaseConverter` interface.
    """

    pass



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.transfer` --
Learning search spaces for Bayesian optimization:
Another view of hyperparameter transfer learning
======================================================================

.. module:: transfer
   :platform: Unix
   :synopsis: Executes transfer optimization of learning search space.

"""
import logging

import numpy as np


log = logging.getLogger(__name__)
# It should apply experiment and trials objects (in its same root path?)
# to represent current task (experiment) and historical tasks (trials)
# and update the space domain in the experiment object
# The data structure is aligned with the object of experiments and trials
# curr_task = {"space":
#                   [{"lr": [lower, upper]},
#                    ...,
#                    {"wd": [lower, upper]}]}

# hist_task = [{"params":
#               [{"name": "lr", "type": real, "value": 0.05},
#                ...,
#                {"name": "wd", "type": real, "value": 0.0001}]},
#              ...,
#              {"params":
#               [{"name": "lr", "type": real, "value": 0.02},
#                ...,
#                {"name": "wd", "type": real, "value": 0.0003}]}]

# constraints: 1) hyper-params are the same in curr and hist;
#              2) all types are real


def rep_task(hist_tasks, curr_task):
    """Represent search space as matrix"""
    hpp_idx_map = {}
    # idx_hpp_map = {}
    curr_bound = np.zeros((2, len(curr_task["space"])))  # 0: lower, 1: upper
    hist_opt = np.zeros((len(hist_tasks), len(curr_task["space"])))

    for i in range(len(curr_task["space"])):
        value_item = next(iter(curr_task["space"][i].values()))
        key_item = next(iter(curr_task["space"][i].keys()))
        hpp_idx_map[key_item] = i
        # idx_hpp_map[i] = key_item
        curr_bound[0, i] = value_item[0]
        curr_bound[1, i] = value_item[1]

    for i in range(len(hist_tasks)):
        for j in range(len(hist_tasks[i]["params"])):
            hist_opt[i, hpp_idx_map[hist_tasks[i]["params"][j]["name"]]] = \
                hist_tasks[i]["params"][j]["value"]

    return hist_opt, curr_bound


def shrink_bbox_space(hist_tasks, curr_task):
    """Shrink search space with bounding box representation according to the historical knowledge"""
    hist_bound = np.zeros((2, len(curr_task["space"])))
    adj_bound = np.zeros((2, len(curr_task["space"])))

    hist_opt, curr_bound = rep_task(hist_tasks, curr_task)

    hist_bound[0, :] = np.min(hist_opt, axis=0)
    hist_bound[1, :] = np.max(hist_opt, axis=0)

    adj_bound[0, :] = np.max(np.stack((hist_bound[0], curr_bound[0])), axis=0)
    adj_bound[1, :] = np.min(np.stack((hist_bound[1], curr_bound[1])), axis=0)

    if np.any(adj_bound[0] > adj_bound[1]):
        return  # if no intersection, reture curr setting space
    for i in range(len(curr_task["space"])):
        key_item = next(iter(curr_task["space"][i].keys()))
        curr_task["space"][i][key_item][0] = adj_bound[0, i]
        curr_task["space"][i][key_item][1] = adj_bound[1, i]
    return


def shrink_ellips_space(hist_tasks, curr_task):
    """Shrink search space with ellipsoidal representation according to the historical knowledge"""
    hyper_param_dim = len(curr_task["space"])
    hist_opt, curr_bound = rep_task(hist_tasks, curr_task)
    matrix_a = np.zeros((hyper_param_dim, hyper_param_dim))
    vector_b = np.zeros((1, hyper_param_dim))

    # need to implement interior-points algorithm to solve the convex problem by CVXPY in future#

    return matrix_a, vector_b


def rej_sampling_ellips_space(matrix_a, vector_b, curr_task):
    """Reject sample from ellipsoidal search space example"""
    hyper_param_dim = vector_b.shape[0]
    _, curr_bound = rep_task([], curr_task)

    is_feasible = False
    while not is_feasible:
        mean = np.zeros(hyper_param_dim)
        cov = np.eye(hyper_param_dim)
        z = np.random.multivariate_normal(mean, cov)
        r = np.random.random()
        scale = np.power(r, 1.0 / hyper_param_dim) / np.linalg.norm(z)
        t = scale * z
        x = np.mat(np.linalg.inv(matrix_a)) * np.matrix.transpose(np.mat(t - vector_b))
        x = np.array(x).ravel()
        diff = curr_bound - np.tile(x, (2, 1))
        if np.all(diff[0, :] <= 0) and np.all(diff[1, :] >= 0):  # x in curr search space
            is_feasible = True

    # x is in the same order as curr_task representation
    return x


if __name__ == "__main__":
    # curr_task = {"space":
    #                   [{"lr": [0.001, 0.04]},
    #                    {"wd": [0.00001, 0.00005]}]}

    curr_task = {"space":
                 [{"lr": [0.001, 0.04]},
                  {"wd": [0.00001, 0.01]}]}

    hist_task = [{"params":
                  [{"name": "lr", "type": "real", "value": 0.05},
                   {"name": "wd", "type": "real", "value": 0.0001}]},
                 {"params":
                  [{"name": "lr", "type": "real", "value": 0.02},
                   {"name": "wd", "type": "real", "value": 0.0003}]},
                 {"params":
                  [{"name": "lr", "type": "real", "value": 0.007},
                   {"name": "wd", "type": "real", "value": 0.0002}]}]

    shrink_bbox_space(hist_task, curr_task)
    print("update curr task:", curr_task)



#!/usr/bin/env python
"""Common fixtures and utils for evc unit-tests."""
import copy

import pytest

from orion.core.evc import conflicts


@pytest.fixture
def parent_config():
    """Generate a new experiment configuration"""
    return dict(
        _id='test',
        name='test',
        metadata={'user': 'corneauf'},
        version=1)


@pytest.fixture
def child_config(parent_config):
    """Generate a new experiment configuration"""
    config = copy.deepcopy(parent_config)
    config['_id'] = "test2"
    config['refers'] = {'parent_id': 'test'}
    config['version'] = 2

    return config


@pytest.fixture
def exp_no_child_conflict(create_db_instance, parent_config):
    """Generate an experiment name conflict"""
    create_db_instance.write('experiments', parent_config)
    return conflicts.ExperimentNameConflict(parent_config, parent_config)


@pytest.fixture
def exp_w_child_conflict(create_db_instance, parent_config, child_config):
    """Generate an experiment name conflict"""
    create_db_instance.write('experiments', parent_config)
    create_db_instance.write('experiments', child_config)
    return conflicts.ExperimentNameConflict(child_config, child_config)


@pytest.fixture
def exp_w_child_as_parent_conflict(create_db_instance, parent_config, child_config):
    """Generate an experiment name conflict"""
    create_db_instance.write('experiments', parent_config)
    create_db_instance.write('experiments', child_config)
    return conflicts.ExperimentNameConflict(parent_config, parent_config)


@pytest.fixture
def existing_exp_conflict(create_db_instance, parent_config):
    """Generate an experiment name conflict"""
    create_db_instance.write('experiments', parent_config)
    create_db_instance.write('experiments', {'name': 'dummy', 'version': 1})
    return conflicts.ExperimentNameConflict(parent_config, parent_config)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.db.set` -- Module running the set command
==============================================================

.. module:: setup
   :platform: Unix
   :synopsis: Update data of experiments and trials in the storage
"""
import argparse
import logging
import sys

import orion.core.io.experiment_builder as experiment_builder
from orion.core.utils.pptree import print_tree
from orion.core.utils.terminal import confirm_name
from orion.storage.base import get_storage


logger = logging.getLogger(__name__)


DESCRIPTION = """
Command to update trial attributes.

To change a trial status, simply give the experiment name,
trial id and status. (use `orion status --all` to get trial ids)
$ orion db set my-exp-name id=3cc91e851e13281ca2152c19d888e937 status=interrupted

To change all trials from a given status to another, simply give the two status
$ orion db set my-exp-name status=broken status=interrupted

Or `*` to apply the change to all trials
$ orion db set my-exp-name '*' status=interrupted

By default, trials of the last version of the experiment are selected.
Add --version to select a prior version. Note that the modification
is applied recursively to all child experiment, but not to the parents.
$ orion db set my-exp-name --version 1 status=broken status=interrupted
"""


CONFIRM_MESSAGE = """
Trials matching the query `{query}` for all experiments listed above
will be modified with `{update}`.
To select a specific version use --version <VERSION>.

Make sure to stop any worker currently executing one of these experiment.

To proceed, type again the name of the experiment: """


def add_subparser(parser):
    """Return the parser that needs to be used for this command"""
    set_parser = parser.add_parser(
        'set',
        description=DESCRIPTION,
        help="Update trials' attributes",
        formatter_class=argparse.RawTextHelpFormatter)

    set_parser.set_defaults(func=main)

    set_parser.add_argument(
        'name',
        help='Name of the experiment to delete.')

    set_parser.add_argument(
        '-c', '--config', type=argparse.FileType('r'), metavar='path-to-config',
        help="user provided orion configuration file")

    set_parser.add_argument(
        '-v', '--version', type=int, default=None,
        help="specific version of experiment to fetch; "
             "(default: last version matching.)")

    set_parser.add_argument(
        '-f', '--force', action='store_true',
        help='Force modify without asking to enter experiment name twice.')

    set_parser.add_argument(
        'query',
        help=(f'Query for trials to update. Can be `*` to update all trials. '
              f'Otherwise format must be <attribute>=<value>. Ex: status=broken. '
              f'Supported attributes are {VALID_QUERY_ATTRS}'))

    set_parser.add_argument(
        'update',
        help=(f'Update for trials. Format must be <attribute>=<value>. '
              f'Ex: status=interrupted. '
              f'Supported attributes are {VALID_UPDATE_ATTRS}'))

    return set_parser


def process_updates(storage, root, query, update):
    """Update the matching trials of the given experiment and its children."""
    trials_total = 0
    for node in root:
        count = storage.update_trials(node.item, where=query, **update)
        logger.debug('%d trials modified in experiment %s-v%d',
                     count, node.item.name, node.item.version)
        trials_total += count

    print(f'{trials_total} trials modified')


VALID_QUERY_ATTRS = ['status', 'id']
VALID_UPDATE_ATTRS = ['status']


def build_query(query):
    """Convert query string to dict format

    String format must be <attr name>=<value>
    """
    if query.strip() == '*':
        return {}

    attribute, value = query.split('=')

    if attribute not in VALID_QUERY_ATTRS:
        raise ValueError(
            f'Invalid query attribute `{attribute}`. Must be one of {VALID_QUERY_ATTRS}')

    if attribute == 'id':
        attribute = '_id'

    return {attribute: value}


def build_update(update):
    """Convert update string to dict format

    String format must be <attr name>=<value>
    """
    attribute, value = update.split('=')

    if attribute not in VALID_UPDATE_ATTRS:
        raise ValueError(
            f'Invalid update attribute `{attribute}`. Must be one of {VALID_UPDATE_ATTRS}')

    return {attribute: value}


def main(args):
    """Remove the experiment(s) or trial(s)."""
    config = experiment_builder.get_cmd_config(args)
    experiment_builder.setup_storage(config.get('storage'))

    # Find root experiment
    root = experiment_builder.build_view(name=args['name'],
                                         version=args.get('version', None)).node

    try:
        query = build_query(args['query'])
        update = build_update(args['update'])
    except ValueError as e:
        print(f'Error: {e}', file=sys.stderr)
        return 1

    # List all experiments with children
    print_tree(root, nameattr='tree_name')

    confirmed = confirm_name(
        CONFIRM_MESSAGE.format(query=args['query'], update=args['update']),
        args['name'], args['force'])

    if not confirmed:
        print('Confirmation failed, aborting operation.')
        return 1

    storage = get_storage()

    process_updates(storage, root, query, update)

    return 0



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.io.database.ephemeraldb` -- Non permanent database
===================================================================

.. module:: database
   :platform: Unix
   :synopsis: Implement non permanent version of :class:`orion.core.io.database.AbstractDB`

"""
from collections import defaultdict
import copy

from orion.core.io.database import AbstractDB, DatabaseError, DuplicateKeyError
from orion.core.utils.flatten import flatten, unflatten


def _convert_keys_to_name(keys):
    index = []

    if len(keys) == 1 and keys[0] == '_id':
        index = '_id_'
    else:
        index = '_'.join('{}_1'.format(k) for k in keys)

    return index


# pylint: disable=too-many-public-methods
class EphemeralDB(AbstractDB):
    """Non permanent database

    This database is meant for debugging purposes. It only lives through one execution and all
    information saved during it is lost when the process is terminated.

    .. seealso:: :class:`orion.core.io.database.AbstractDB` for more on attributes.

    """

    @property
    def is_connected(self):
        """Return true, always."""
        return True

    def initiate_connection(self):
        """Create the dictionary which serve as an ephemeral database"""
        self._db = defaultdict(EphemeralCollection)

    def close_connection(self):
        """Remove the dictionary"""
        self._db = None

    def ensure_index(self, collection_name, keys, unique=False):
        """Create given indexes if they do not already exist in database.

        Indexes are only created if `unique` is True.
        """
        self._db[collection_name].create_index(keys, unique=unique)

    def index_information(self, collection_name):
        """Return dict of names and sorting order of indexes"""
        return self._db[collection_name].index_information()

    def drop_index(self, collection_name, name):
        """Remove index from the database"""
        self._db[collection_name].drop_index(name)

    def write(self, collection_name, data, query=None):
        """Write new information to a collection. Perform insert or update.

        .. seealso:: :meth:`AbstractDB.write` for argument documentation.

        """
        dbcollection = self._db[collection_name]

        if query is None:
            # We can assume that we do not want to update.
            # So we do insert_many instead.
            if type(data) not in (list, tuple):
                data = [data]
            return dbcollection.insert_many(documents=data)

        update_data = {'$set': data}

        return dbcollection.update_many(query=query,
                                        update=update_data)

    def read(self, collection_name, query=None, selection=None):
        """Read a collection and return a value according to the query.

        .. seealso:: :meth:`AbstractDB.read` for argument documentation.

        """
        dbcollection = self._db[collection_name]

        dbdocs = dbcollection.find(query, selection)

        return dbdocs

    def read_and_write(self, collection_name, query, data, selection=None):
        """Read a collection's document and update the found document.

        Returns the updated document, or None if nothing found.

        .. seealso:: :meth:`AbstractDB.read_and_write` for
                     argument documentation.

        """
        dbdoc = self.read(collection_name, query)
        if not dbdoc:
            return None

        id_query = {'_id': dbdoc[0]['_id']}
        self.write(collection_name, data, id_query)
        return self.read(collection_name, id_query)[0]

    def count(self, collection_name, query=None):
        """Count the number of documents in a collection which match the `query`.

        .. seealso:: :meth:`AbstractDB.count` for argument documentation.

        """
        dbcollection = self._db[collection_name]
        return dbcollection.count(query=query)

    def remove(self, collection_name, query):
        """Delete from a collection document[s] which match the `query`.

        .. seealso:: :meth:`AbstractDB.remove` for argument documentation.

        """
        dbcollection = self._db[collection_name]

        return dbcollection.delete_many(query=query)


class EphemeralCollection(object):
    """Non permanent collection

    This collection is meant for debugging purposes within the EphemeralDB.

    .. seealso:: :class:`orion.core.io.database.ephemeraldb.EphemeralDB` for database object.

    """

    def __init__(self):
        """Initialise the collection, with no documents and only _id unique index."""
        self._documents = []
        self._indexes = dict()
        self.create_index('_id', unique=True)

    def create_index(self, keys, unique=False):
        """Create given indexes if they do not already exist for this collection.

        Indexes are only created if `unique` is True.
        """
        # turn single key into list for coherence
        if not isinstance(keys, (list, tuple)):
            keys = [(keys, None)]

        keys = tuple(key for (key, order) in keys)
        name = _convert_keys_to_name(keys)
        if unique and name not in self._indexes:
            data = set()

            self._indexes[name] = (keys, data)

            for document in self._documents:
                self._validate_index(document, indexes=[name])
                data.add(tuple(document[key] for key in keys))

    def index_information(self):
        """Return dict of names and sorting order of indexes

        EphemeralCollection may only contain unique indexes.
        """
        return {name: True for name in self._indexes}

    def drop_index(self, name):
        """Remove index from the database

        EphemeralCollection may only contain unique indexes.
        """
        if name not in self._indexes:
            raise DatabaseError('index not found with name {}'.format(name))

        del self._indexes[name]

    def _register_keys(self, document):
        """Register index values of a new document"""
        for keys, values in self._indexes.values():
            values.add(tuple(document[key] for key in keys))

    def find(self, query=None, selection=None):
        """Find documents in the collection and return a value according to the query.

        .. seealso:: :meth:`AbstractDB.read` for argument documentation.

        """
        found_documents = []
        for document in self._documents:
            if document.match(query):
                found_documents.append(document.select(selection))

        return found_documents

    def _validate_index(self, document, indexes=None):
        """Validate index values of a document

        Raises
        ------
        DuplicateKeyError
            If the document contains unique indexes which are already present in the database.

        """
        if indexes is None:
            indexes = self._indexes.keys()

        for name in indexes:
            keys, data = self._indexes[name]
            document_values = tuple(document[key] for key in keys)
            if document_values in data:
                raise DuplicateKeyError(
                    "Duplicate key error: index={} value={}".format(name, document_values))

    def _get_new_id(self):
        """Return max id + 1"""
        if self._documents:
            return max(d['_id'] for d in self._documents) + 1

        return 1

    def insert_many(self, documents):
        """Add new documents in the collection.

        If the documents do not have a keys `_id`, they are assigned by default
        the max id + 1.

        Raises
        ------
        DuplicateKeyError
            If the document contains unique indexes which are already present in the database.

        """
        for document in documents:
            if '_id' not in document:
                document['_id'] = self._get_new_id()
            ephemeral_document = EphemeralDocument(document)
            self._validate_index(ephemeral_document)
            self._documents.append(ephemeral_document)
            self._register_keys(ephemeral_document)

        return len(documents)

    def update_many(self, query, update):
        """Update documents matching the query

        Raises
        ------
        DuplicateKeyError
            If the update creates a duplication of unique indexes in the database.

        """
        updates = 0
        for document in self._documents:
            if document.match(query):
                document.update(update)
                updates += 1

        return updates

    def _upsert(self, query, update):
        """Insert the document when query was not found.

        If update contains `$set`, then the new document is the combination of query and
        update['$set'], otherwise the new document is `update`.
        """
        if "$set" in update:
            new_document = copy.deepcopy(query)
            new_document.update(update["$set"])
        else:
            new_document = update

        self.insert_many([new_document])

    def count(self, query=None):
        """Count the number of documents in a collection which match the `query`.

        .. seealso:: :meth:`AbstractDB.count` for argument documentation.

        """
        return len(self.find(query))

    def delete_many(self, query=None):
        """Delete from a collection document[s] which match the `query`.

        .. seealso:: :meth:`AbstractDB.remove` for argument documentation.

        """
        deleted = 0
        retained_documents = []
        for document in self._documents:
            if not document.match(query):
                retained_documents.append(document.to_dict())
            else:
                deleted += 1

        # Reset indexes
        for name, (keys, _) in self._indexes.items():
            self._indexes[name] = (keys, set())

        self._documents = []
        self.insert_many(retained_documents)

        return deleted

    def drop(self):
        """Drop the collection, removing all documents and indexes."""
        self._documents = []
        self._indexes = dict()
        self.create_index('_id', unique=True)


class EphemeralDocument(object):
    """Non permanent document

    This document is meant for debugging purposes within the EphemeralDB.

    .. seealso:: :class:`orion.core.io.database.ephemeraldb.EphemeralDB` for database object.

    """

    operators = {
        "$ne": (lambda a, b: a != b),
        "$in": (lambda a, b: a in b),
        "$gte": (lambda a, b: a is not None and a >= b),
        "$gt": (lambda a, b: a is not None and a > b),
        "$lte": (lambda a, b: a is not None and a <= b),
    }

    def __init__(self, data):
        """Initialise the document with a flattened version of the data"""
        self._data = flatten(data)

    def match(self, query=None):
        """Test if the document corresponds to a given query"""
        if query is None or query == {}:
            return True

        query = flatten(query)
        for key, value in query.items():
            if not self.match_key(key, value):
                return False

        return True

    def _is_operator(self, key):  # pylint: disable=no-self-use
        return key.split(".")[-1].startswith('$')

    def _get_key_operator(self, key):
        path = key.split(".")
        operator = path[-1]
        key = ".".join(path[:-1])

        if operator not in self.operators:
            raise ValueError('Operator \'{}\' is not supported by EphemeralDB'.format(operator))

        return key, self.operators[operator]

    def match_key(self, key, value):
        """Test if a data corresponding to the given key is in agreement with the given
        value based on the operator defined within the key.

        Default operator is equal when no operator is defined.
        Other operators could be $ne, $in, $gte, $gt or $lte. They are defined
        in the last section of the key. For example: `abc.def.$in` or `abc.def.$gte`.
        """
        if self._is_operator(key):
            key, operator = self._get_key_operator(key)

            return key in self and operator(self[key], value)

        return key in self and self[key] == value

    def _validate_keys(self, keys):
        """Verify that all keys are 0 or 1 (with exception of _id) and convert them.

        For simplicity, when keys are 0, the inverse set of keys for 1s is computed.

        .. note ::

            _id is set to 1 if not specified. Only _id may be set to 0 if other keys are set to 1.
        """
        if len(keys) == 1 and keys.get('_id', 0) == 1:
            return keys

        keys_without_id = [key for key in keys if key != '_id']
        n_keys = sum(keys[key] for key in keys_without_id)
        if n_keys != 0 and n_keys != len(keys_without_id):
            raise ValueError(
                'Cannot mix selection with 1 and 0s except for _id: {}'.format(keys))

        # All given keys are 0 (with possible exception of _id)
        if n_keys == 0:
            new_keys = dict((key, 1) for key in self._data.keys() if key not in keys)
            new_keys['_id'] = keys.get('_id', 1)
            keys = new_keys

        keys.setdefault('_id', 1)

        return keys

    def select(self, keys):
        """Only select or only drop the specified keys

        For a pair (key, value) in the dictionnary, value=0 means the key will not be included
        while value=1 means it will.

        All specified keys should be 0 or 1. They cannot have different values with the exception
        of _id which can be specified to 0 while the others are at 1. The _id field is always
        returned unless specified with 0.

        Parameters
        ----------
        keys: dict
            Pairs of keys and 0 or 1s. When a key is associated with 1, it is kept in the selection,
            otherwise it is dropped.

        """
        if not keys:
            return unflatten(self._data)

        keys = flatten(keys)
        keys = self._validate_keys(keys)

        selection = dict()

        def key_is_match(key, selected_key):
            """Test if key matches the selected key

            key_is_match(abc.def.ghi, abc.def.ghi) -> True
            key_is_match(abc.def.ghi, abc.def) -> True
            key_is_match(abc.def.ghi, abc.de) -> False
            key_is_match(abc.def.ghi, xyz) -> False
            """
            return (key == selected_key or
                    (key.startswith(selected_key) and
                     key.replace(selected_key, '')[0] == "."))

        for selected_key, include in filter(lambda item: item[1], keys.items()):
            match = False
            for key, value in self._data.items():
                if include and key_is_match(key, selected_key):
                    match = True
                    selection[key] = value

            if not match:
                selection[selected_key] = None

        return unflatten(selection)

    def update(self, data):
        """Update the values of the document.

        Parameters
        ----------
        data: dict
            Dictionary of data to update the document. If `$set` is in
            the data, the corresponding `data[$set]` will be used instead.

        """
        data = flatten(data.get("$set", data))
        self._data.update(data)

    def to_dict(self):
        """Convert the ephemeral document to a python dictionary"""
        return self.select({})

    def __getitem__(self, key):
        """Get the item corresponding to the given key in the document"""
        return self._data.get(key, None)

    def __contains__(self, key):
        """Test whether the given key is present in the document"""
        return key in self._data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.worker.trials_history`."""

from orion.core.worker.trials_history import TrialsHistory


class DummyTrial(object):
    """Mocking class for the Trial"""

    def __init__(self, trial_id, parents):
        """Init _id and parents only"""
        self.id = trial_id
        self.parents = parents


def test_history_contains_new_child():
    """Verify that __contains__ return True for a new child"""
    trials_history = TrialsHistory()
    new_child = DummyTrial(1, [])
    assert new_child not in trials_history
    trials_history.update([new_child])
    assert new_child in trials_history


def test_history_contains_old_child():
    """Verify that __contains__ return True for a new child"""
    trials_history = TrialsHistory()
    old_child = DummyTrial(1, [])
    trials_history.update([old_child])
    new_child = DummyTrial(2, [old_child.id])
    assert new_child not in trials_history
    trials_history.update([new_child])
    assert old_child.id not in trials_history.children
    assert old_child in trials_history
    assert new_child.id in trials_history.children
    assert new_child in trials_history


def test_added_children_without_ancestors():
    """Verify that children are added to history"""
    trials_history = TrialsHistory()
    trials_history.update([DummyTrial(i, []) for i in range(3)])
    assert trials_history.children == [0, 1, 2]
    trials_history.update([DummyTrial(i, []) for i in range(3, 6)])
    assert trials_history.children == [0, 1, 2, 3, 4, 5]


def test_added_children_with_ancestors():
    """Verify that children with ancestors are added to history"""
    trials_history = TrialsHistory()
    trials = [DummyTrial(i, []) for i in range(3)]
    trials_history.update(trials)
    assert trials_history.children == [0, 1, 2]

    trials = [DummyTrial(i, [trials[i % 3].id]) for i in range(3, 6)]
    trials_history.update(trials)
    assert len(set(trials_history.children) & set([3, 4, 5])) == 3


def test_discarded_children():
    """Verify that ancestors of new children are discarded from history"""
    trials_history = TrialsHistory()
    trials = [DummyTrial(i, []) for i in range(3)]
    trials_history.update(trials)
    assert trials_history.children == [0, 1, 2]

    trials_history.update(trials)
    assert trials_history.children == [0, 1, 2]

    trials = [DummyTrial(i, [trials[i % 3].id]) for i in range(3, 6)]
    trials_history.update(trials)
    assert trials_history.children == [3, 4, 5]


def test_discarded_duplicate_children():
    """Verify that duplicate children are not added twice"""
    trials_history = TrialsHistory()
    trials = [DummyTrial(i, []) for i in range(3)]
    trials_history.update(trials)
    assert trials_history.children == [0, 1, 2]

    trials = [DummyTrial(i, [trials[i].id]) for i in range(3)]
    assert all(trial.id == trial.parents[0] for trial in trials)
    trials_history.update(trials)
    assert trials_history.children == [0, 1, 2]


def test_discarded_shared_children():
    """Verify that only ancestors are removed and not all past children"""
    trials_history = TrialsHistory()
    trials = [DummyTrial(i, []) for i in range(3)]
    trials_history.update(trials)
    assert trials_history.children == [0, 1, 2]

    trials = [DummyTrial(i, [0]) for i in range(3, 6)]
    trials_history.update(trials)
    assert trials_history.children == [1, 2, 3, 4, 5]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Oríon documentation build configuration file.

This file is execfile()d with the current directory set to its
containing dir.

Note that not all possible configuration values are present in this
autogenerated file.

All configuration values have a default; values that are commented out
serve to show the default.

If extensions (or modules to document with autodoc) are in another directory,
add these directories to sys.path here. If the directory is relative to the
documentation root, use os.path.abspath to make it absolute, like shown here.

How to document -- sources:

   1. `Numpy Standard <https://numpydoc.readthedocs.io/en/latest/format.html>`_
   2. `Python Standard <https://docs.python.org/devguide/documenting.html>`_
   3. `reST general <http://www.sphinx-doc.org/en/stable/rest.html>`_
   4. `reST reference tags <http://www.sphinx-doc.org/en/stable/domains.html#the-python-domain>`_
   5. `Cross-reference <http://www.sphinx-doc.org/en/stable/domains.html#python-roles>`_

"""
import glob
import os
import re
import sys

docs_src_path = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, docs_src_path)
src_path = os.path.abspath(os.path.join(docs_src_path, '..', '..', 'src'))
sys.path.insert(0, src_path)

import orion.core as orion  # noqa

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

extensions = [  # Extensions must be sorted alphabetically to ease maintenance and merges.
    'numpydoc',
    'sphinxcontrib.httpdomain',  # Documentation directives for the REST API.
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.extlinks',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode'
]

# General information about the project.
project = u'orion'
_full_version = orion.__version__
author = orion.__author__
copyright = orion.__copyright__

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = re.sub(r'(.*?)(?:\.dev\d+)?(?:\+.*)?', r'\1', _full_version)
# The short X.Y version.
version = re.sub(r'(\d+)(\.\d+)?(?:\.\d+)?(?:-.*)?(?:\.post\d+)?', r'\1\2', release)

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', '**/pptree.py']

# The name of the Pygments (syntax highlighting) style to use.
highlight_language = 'python3'
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

# The reST default role (used for this markup: `text`) to use for all documents.
default_role = "autolink"

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    'canonical_url': '',
    'analytics_id': '',
    'logo_only': False,
    'display_version': True,
    'prev_next_buttons_location': 'both',
    #  'style_external_links': False,
    #  'vcs_pageview_mode': '',
    # Toc options
    'collapse_navigation': False,
    'sticky_navigation': True,
    'navigation_depth': 4,
    #  'includehidden': False,
    #  'titles_only': False
}

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# This is required for the alabaster theme
# refs: http://alabaster.readthedocs.io/en/latest/installation.html#sidebars
html_sidebars = {
    '**': [
        'relations.html',  # needs 'show_related': True theme option to display
        'searchbox.html',
    ]
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'oriondoc'


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'orion.tex', u'Oríon Documentation',
     u'Epistímio', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'orion', 'Oríon Documentation',
     [author], 1)
]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'Oríon', 'Oríon Documentation',
     author, 'orion', orion.__descr__,
     'Miscellaneous'),
]

# -- Autodoc configuration -----------------------------------------------

autodoc_mock_imports = ['_version', 'utils._appdirs']

################################################################################
#                             Numpy Doc Extension                              #
################################################################################

# sphinx.ext.autosummary will automatically be loaded as well. So:
autosummary_generate = glob.glob("reference/*.rst")

# Generate ``plot::`` for ``Examples`` sections which contain matplotlib
numpydoc_use_plots = False

# Create a Sphinx table of contents for the lists of class methods and
# attributes. If a table of contents is made, Sphinx expects each entry to have
# a separate page.
numpydoc_class_members_toctree = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Installation script for Oríon."""
import os

from setuptools import setup

import versioneer


repo_root = os.path.dirname(os.path.abspath(__file__))


tests_require = [
    'pytest>=3.0.0'
    'scikit-learn'
    ]


packages = [  # Packages must be sorted alphabetically to ease maintenance and merges.
    'orion.algo',
    'orion.analysis',
    'orion.client',
    'orion.core',
    'orion.plotting',
    'orion.serving',
    'orion.storage',
    'orion.testing'
    ]

setup_args = dict(
    name='orion',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Asynchronous [black-box] Optimization',
    long_description=open(os.path.join(repo_root, "README.rst"), 'rt', encoding='utf8').read(),
    license='BSD-3-Clause',
    author=u'Epistímio',
    author_email='xavier.bouthillier@umontreal.ca',
    url='https://github.com/epistimio/orion',
    packages=packages,
    package_dir={'': 'src'},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'orion = orion.core.cli:main',
            ],
        'OptimizationAlgorithm': [
            'random = orion.algo.random:Random',
            'asha = orion.algo.asha:ASHA',
            'hyperband = orion.algo.hyperband:Hyperband',
            'tpe = orion.algo.tpe:TPE',
            'EvolutionES = orion.algo.evolution_es:EvolutionES',
            ],
        'Storage': [
            'track = orion.storage.track:Track',
            'legacy = orion.storage.legacy:Legacy',
            ]
        },
    install_requires=['PyYAML', 'pymongo>=3', 'numpy', 'scipy', 'gitpython', 'filelock',
                      'tabulate', 'AppDirs', 'plotly', 'pandas', 'gunicorn', 'falcon'],
    tests_require=tests_require,
    setup_requires=['setuptools', 'appdirs', 'pytest-runner'],
    extras_require=dict(test=tests_require),
    # "Zipped eggs don't play nicely with namespace packaging"
    # from https://github.com/pypa/sample-namespace-packages
    zip_safe=False
    )

setup_args['keywords'] = [
    'Machine Learning',
    'Deep Learning',
    'Distributed',
    'Optimization',
    ]

setup_args['platforms'] = ['Linux']

setup_args['classifiers'] = [
    'Development Status :: 1 - Planning',
    'Intended Audience :: Developers',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Programming Language :: Python',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
] + [('Programming Language :: Python :: %s' % x)
     for x in '3 3.6 3.7 3.8'.split()]

if __name__ == '__main__':
    setup(**setup_args)



# -*- coding: utf-8 -*-
"""
:mod:`orion.algo.gradient_descent` -- Perform gradient descent on a loss surface
================================================================================

.. module:: gradient_descent
   :platform: Unix
   :synopsis: Use gradients to locally search for a minimum.

"""
import numpy

from orion.algo.base import BaseAlgorithm


class Gradient_Descent(BaseAlgorithm):
    """Implement a gradient descent algorithm."""

    requires = 'real'

    def __init__(self, space, learning_rate=1., dx_tolerance=1e-7):
        """Declare `learning_rate` as a hyperparameter of this algorithm."""
        super(Gradient_Descent, self).__init__(space,
                                               learning_rate=learning_rate,
                                               dx_tolerance=dx_tolerance)
        self.has_observed_once = False
        self.current_point = None
        self.gradient = numpy.array([numpy.inf])

    def suggest(self, num=1):
        """Suggest a `num`ber of new sets of parameters.

        Perform a step towards negative gradient and suggest that point.

        """
        assert num == 1  # Simple gradient descent does not make sense else
        if not self.has_observed_once:
            return self.space.sample(1)

        self.current_point -= self.learning_rate * self.gradient
        return [self.current_point]

    def observe(self, points, results):
        """Observe evaluation `results` corresponding to list of `points` in
        space.

        Save current point and gradient corresponding to this point.

        """
        self.current_point = numpy.asarray(points[-1])
        self.gradient = numpy.asarray(results[-1]['gradient'])
        self.has_observed_once = True

    @property
    def is_done(self):
        """Implement a terminating condition."""
        dx = self.learning_rate * numpy.sqrt(self.gradient.dot(self.gradient))
        return dx <= self.dx_tolerance



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.evc` -- Experiment Version Control System
==========================================================

.. module:: evc
   :platform: Unix
   :synopsis: Organize related experiments inside a tree-structure using adapters to maintain
      compatibility between them.
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.db.upgrade` -- Module to upgrade DB schemes
================================================================

.. module:: test_db
   :platform: Unix
   :synopsis: Upgrade the scheme of the databases

"""
import argparse
import logging
import sys

from orion.core.io.database.ephemeraldb import EphemeralCollection
from orion.core.io.database.mongodb import MongoDB
from orion.core.io.database.pickleddb import PickledDB
import orion.core.io.experiment_builder as experiment_builder
import orion.core.utils.backward as backward
from orion.storage.base import get_storage
from orion.storage.legacy import Legacy


log = logging.getLogger(__name__)
SHORT_DESCRIPTION = 'Upgrade the database scheme'


# TODO: Move somewhere else to share with `db setup`.
def ask_question(question, default=None):
    """Ask a question to the user and receive an answer.

    Parameters
    ----------
    question: str
        The question to be asked.
    default: str
        The default value to use if the user enters nothing.

    Returns
    -------
    str
        The answer provided by the user.

    """
    if default is not None:
        question = question + " (default: {}) ".format(default)

    answer = input(question)

    if answer.strip() == "":
        return default

    return answer


def add_subparser(parser):
    """Add the subparser that needs to be used for this command"""
    upgrade_db_parser = parser.add_parser('upgrade', help=SHORT_DESCRIPTION,
                                          description=SHORT_DESCRIPTION)

    upgrade_db_parser.add_argument('-c', '--config', type=argparse.FileType('r'),
                                   metavar='path-to-config', help="user provided "
                                   "orion configuration file")

    upgrade_db_parser.add_argument('-f', '--force', action='store_true',
                                   help="Don't prompt user")

    upgrade_db_parser.set_defaults(func=main)

    return upgrade_db_parser


def main(args):
    """Upgrade the databases for current version"""
    print("Upgrading your database may damage your data. Make sure to make a backup before the "
          "upgrade and stop any other process that may read/write the database during the upgrade.")

    if not args.get('force'):
        action = ''
        while action not in ['y', 'yes', 'no', 'n']:
            action = ask_question("Do you wish to proceed? (y/N)", "N").lower()

        if action in ['no', 'n']:
            sys.exit(0)

    config = experiment_builder.get_cmd_config(args)
    storage_config = config.get('storage')

    if storage_config is None:
        storage_config = {'type': 'legacy'}

    storage_config['setup'] = False

    experiment_builder.setup_storage(storage_config)

    storage = get_storage()

    upgrade_db_specifics(storage)

    print('Updating documents...')
    upgrade_documents(storage)
    print('Database upgrade completed successfully')


def upgrade_db_specifics(storage):
    """Make upgrades that are specific to some backends"""
    if isinstance(storage, Legacy):
        database = storage._db  # pylint: disable=protected-access
        print('Updating indexes...')
        update_indexes(database)
        if isinstance(database, PickledDB):
            print('Updating pickledb scheme...')
            upgrade_pickledb(database)
        elif isinstance(database, MongoDB):
            print('Updating mongodb scheme...')
            upgrade_mongodb(database)


def upgrade_documents(storage):
    """Upgrade scheme of the documents"""
    for experiment in storage.fetch_experiments({}):
        add_version(experiment)
        add_space(experiment)
        storage.update_experiment(uid=experiment.pop('_id'), **experiment)


def add_version(experiment):
    """Add version 1 if not present"""
    experiment.setdefault('version', 1)


def add_space(experiment):
    """Add space to metadata if not present"""
    backward.populate_space(experiment)


def update_indexes(database):
    """Remove user from unique indices.

    This is required for migration to v0.1.6+
    """
    # For backward compatibility
    index_info = database.index_information('experiments')
    deprecated_indices = [('name', 'metadata.user'), ('name', 'metadata.user', 'version'),
                          'name_1_metadata.user_1', 'name_1_metadata.user_1_version_1']

    for deprecated_idx in deprecated_indices:
        if deprecated_idx in index_info:
            database.drop_index('experiments', deprecated_idx)


# pylint: disable=unused-argument
def upgrade_mongodb(database):
    """Update mongo specific db scheme."""
    pass


def upgrade_pickledb(database):
    """Update pickledb specific db scheme."""
    # pylint: disable=protected-access
    def upgrade_state(self, state):
        """Set state while ensuring backward compatibility"""
        self._documents = state['_documents']

        # if indexes are from <=v0.1.6
        if state['_indexes'] and isinstance(next(iter(state['_indexes'].keys())), tuple):
            self._indexes = dict()
            for keys, values in state['_indexes'].items():
                if isinstance(keys, str):
                    self._indexes[keys] = values
                # Convert keys that were registered with old index signature
                else:
                    keys = [(key, None) for key in keys]
                    self.create_index(keys, unique=True)
        else:
            self._indexes = state['_indexes']

    old_setstate = getattr(EphemeralCollection, '__setstate__', None)
    EphemeralCollection.__setstate__ = upgrade_state

    document = database.read('experiments', {})[0]
    # One document update is enough to fix all collections
    database.write('experiments', document, query={'_id': document['_id']})

    if old_setstate is not None:
        EphemeralCollection.__setstate__ = old_setstate
    else:
        del EphemeralCollection.__setstate__



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.storage`."""
import copy
import json
import logging
import tempfile

import pytest

from orion.core.io.database import Database
from orion.core.io.database.pickleddb import PickledDB
from orion.core.utils.exceptions import MissingResultFile
from orion.core.utils.singleton import SingletonAlreadyInstantiatedError, \
    SingletonNotInstantiatedError, update_singletons
from orion.core.worker.trial import Trial
from orion.storage.base import FailedUpdate
from orion.storage.legacy import get_database, setup_database
from orion.testing import OrionState

log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)


base_experiment = {
    'name': 'default_name',
    'version': 0,
    'metadata': {
        'user': 'default_user',
        'user_script': 'abc',
        'datetime': '2017-11-23T02:00:00'
    }
}

base_trial = {
    'experiment': 'default_name',
    'status': 'new',  # new, reserved, suspended, completed, broken
    'worker': None,
    'submit_time': '2017-11-23T02:00:00',
    'start_time': None,
    'end_time': None,
    'heartbeat': None,
    'results': [
        {'name': 'loss',
         'type': 'objective',  # objective, constraint
         'value': 2}
    ],
    'params': [
        {'name': '/encoding_layer',
         'type': 'categorical',
         'value': 'rnn'},
        {'name': '/decoding_layer',
         'type': 'categorical',
         'value': 'lstm_with_attention'}
    ]
}


mongodb_config = {
    'database': {
        'type': 'MongoDB',
        'name': 'orion_test',
        'username': 'user',
        'password': 'pass'
    }
}

db_backends = [
    {
        'type': 'legacy',
        'database': mongodb_config
    }
]


@pytest.mark.usefixtures("setup_pickleddb_database")
def test_setup_database_default(monkeypatch):
    """Test that database is setup using default config"""
    update_singletons()
    setup_database()
    database = Database()
    assert isinstance(database, PickledDB)


def test_setup_database_bad():
    """Test how setup fails when configuring with non-existant backends"""
    update_singletons()
    with pytest.raises(NotImplementedError) as exc:
        setup_database({'type': 'idontexist'})

    assert exc.match('idontexist')


def test_setup_database_custom():
    """Test setup with local configuration"""
    update_singletons()
    setup_database({'type': 'pickleddb', 'host': 'test.pkl'})
    database = Database()
    assert isinstance(database, PickledDB)
    assert database.host == 'test.pkl'


def test_setup_database_bad_override():
    """Test setup with different type than existing singleton"""
    update_singletons()
    setup_database({'type': 'pickleddb', 'host': 'test.pkl'})
    database = Database()
    assert isinstance(database, PickledDB)
    with pytest.raises(SingletonAlreadyInstantiatedError) as exc:
        setup_database({'type': 'mongodb'})

    assert exc.match('A singleton instance of \(type: Database\)')


@pytest.mark.xfail(reason='Fix this when introducing #135 in v0.2.0')
def test_setup_database_bad_config_override():
    """Test setup with different config than existing singleton"""
    update_singletons()
    setup_database({'type': 'pickleddb', 'host': 'test.pkl'})
    database = Database()
    assert isinstance(database, PickledDB)
    with pytest.raises(SingletonAlreadyInstantiatedError):
        setup_database({'type': 'pickleddb', 'host': 'other.pkl'})


def test_get_database_uninitiated():
    """Test that get database fails if no database singleton exist"""
    update_singletons()
    with pytest.raises(SingletonNotInstantiatedError) as exc:
        get_database()

    assert exc.match('No singleton instance of \(type: Database\) was created')


def test_get_database():
    """Test that get database gets the singleton"""
    update_singletons()
    setup_database({'type': 'pickleddb', 'host': 'test.pkl'})
    database = get_database()
    assert isinstance(database, PickledDB)
    assert get_database() == database


class TestLegacyStorage:
    """Test Legacy Storage retrieve result mechanic separately"""

    def test_push_trial_results(self, storage=None):
        """Successfully push a completed trial into database."""
        reserved_trial = copy.deepcopy(base_trial)
        reserved_trial['status'] = 'reserved'
        with OrionState(experiments=[], trials=[reserved_trial], storage=storage) as cfg:
            storage = cfg.storage()
            trial = storage.get_trial(Trial(**reserved_trial))
            results = [
                Trial.Result(name='loss', type='objective', value=2)
            ]
            trial.results = results
            assert storage.push_trial_results(trial), 'should update successfully'

            trial2 = storage.get_trial(trial)
            assert trial2.results == results

    def test_push_trial_results_unreserved(self, storage=None):
        """Successfully push a completed trial into database."""
        with OrionState(experiments=[], trials=[base_trial], storage=storage) as cfg:
            storage = cfg.storage()
            trial = storage.get_trial(Trial(**base_trial))
            results = [
                Trial.Result(name='loss', type='objective', value=2)
            ]
            trial.results = results
            with pytest.raises(FailedUpdate):
                storage.push_trial_results(trial)

    def retrieve_result(self, storage, generated_result):
        """Test retrieve result"""
        results_file = tempfile.NamedTemporaryFile(
            mode='w', prefix='results_', suffix='.log', dir='.', delete=True
        )

        # Generate fake result
        json.dump([generated_result], results_file)
        results_file.flush()

        # --
        with OrionState(experiments=[], trials=[], storage=storage) as cfg:
            storage = cfg.storage()

            trial = Trial(**base_trial)
            trial = storage.retrieve_result(trial, results_file)

            results = trial.results

            assert len(results) == 1
            assert results[0].to_dict() == generated_result

        results_file.close()

    def test_retrieve_result(self, storage=None):
        """Test retrieve result"""
        self.retrieve_result(storage, generated_result={
            'name': 'loss',
            'type': 'objective',
            'value': 2})

    def test_retrieve_result_incorrect_value(self, storage=None):
        """Test retrieve result"""
        with pytest.raises(ValueError) as exec:
            self.retrieve_result(storage, generated_result={
                'name': 'loss',
                'type': 'objective_unsupported_type',
                'value': 2})

        assert exec.match(r'Given type, objective_unsupported_type')

    def test_retrieve_result_nofile(self, storage=None):
        """Test retrieve result"""
        results_file = tempfile.NamedTemporaryFile(
            mode='w', prefix='results_', suffix='.log', dir='.', delete=True
        )

        with OrionState(experiments=[], trials=[], storage=storage) as cfg:
            storage = cfg.storage()

            trial = Trial(**base_trial)

            with pytest.raises(MissingResultFile) as exec:
                storage.retrieve_result(trial, results_file)

        results_file.close()

        assert exec.match(r'Cannot parse result file')



# -*- coding: utf-8 -*-
"""
:mod:`orion.serving.trials_resource` -- Module responsible for the trials/ REST endpoint
========================================================================================

.. module:: trials_resource
   :platform: Unix
   :synopsis: Serves all the requests made to trials/ REST endpoint
"""
import json

from falcon import Request, Response

from orion.serving.parameters import retrieve_experiment, retrieve_trial, verify_query_parameters, \
    verify_status
from orion.serving.responses import build_trial_response, build_trials_response
from orion.storage.base import get_storage

SUPPORTED_PARAMETERS = ['ancestors', 'status', 'version']


class TrialsResource(object):
    """Serves all the requests made to trials/ REST endpoint"""

    def __init__(self):
        self.storage = get_storage()

    def on_get_trials_in_experiment(self, req: Request, resp: Response, experiment_name: str):
        """
        Handle GET requests for trials/:experiment where ``experiment`` is
        the user-defined name of the experiment.
        """
        verify_query_parameters(req.params, SUPPORTED_PARAMETERS)

        status = req.get_param('status', default=None)
        verify_status(status)
        version = req.get_param_as_int('version')
        with_ancestors = req.get_param_as_bool('ancestors', default=False)

        experiment = retrieve_experiment(experiment_name, version)
        if status:
            trials = experiment.fetch_trials_by_status(status, with_ancestors)
        else:
            trials = experiment.fetch_trials(with_ancestors)

        response = build_trials_response(trials)
        resp.body = json.dumps(response)

    def on_get_trial_in_experiment(self, req: Request, resp: Response, experiment_name: str,
                                   trial_id: str):
        """
        Handle GET requests for trials/:experiment/:trial_id where ``experiment`` is
        the user-defined name of the experiment and ``trial_id`` the id of the trial.
        """
        experiment = retrieve_experiment(experiment_name)
        trial = retrieve_trial(experiment, trial_id)

        response = build_trial_response(trial)
        resp.body = json.dumps(response)



"""Perform tests for the REST endpoint `/plots`"""
from orion.testing import create_experiment

config = dict(
    name='experiment-name',
    space={'x': 'uniform(0, 200)'},
    metadata={'user': 'test-user',
              'orion_version': 'XYZ',
              'VCS': {"type": "git",
                      "is_dirty": False,
                      "HEAD_sha": "test",
                      "active_branch": None,
                      "diff_sha": "diff"}},
    version=1,
    pool_size=1,
    max_trials=10,
    working_dir='',
    algorithms={'random': {'seed': 1}},
    producer={'strategy': 'NoParallelStrategy'},
)

trial_config = {
    'experiment': 0,
    'status': 'completed',
    'worker': None,
    'start_time': None,
    'end_time': None,
    'heartbeat': None,
    'results': [],
    'params': []
}


def test_root_not_available(client):
    """Tests that plots/regret is not available"""
    response = client.simulate_get('/plots/regret')

    assert response.status == "404 Not Found"


class TestRegretPlots:
    """Tests regret plots"""

    def test_unknown_experiment(self, client):
        """Tests that the API returns a 404 Not Found when an unknown experiment is queried."""
        response = client.simulate_get('/plots/regret/unknown-experiment')

        assert response.status == "404 Not Found"
        assert response.json == {'title': 'Experiment not found',
                                 'description': 'Experiment "unknown-experiment" does not exist'}

    def test_plot(self, client):
        """Tests that the API returns the plot in json format."""
        with create_experiment(config, trial_config, ['completed']) as (_, _, experiment):
            response = client.simulate_get('/plots/regret/experiment-name')

        assert response.status == "200 OK"
        assert response.json
        assert list(response.json.keys()) == ['data', 'layout']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# pylint: disable=eval-used,protected-access
"""
:mod:`orion.core.cli.insert` -- Module to insert new trials
===========================================================

.. module:: insert
   :platform: Unix
   :synopsis: Insert creates new trials for a given experiment with fixed values

"""
import collections
import logging
import os
import re

from orion.core.cli import base as cli
from orion.core.io.convert import infer_converter_from_file_type
import orion.core.io.experiment_builder as experiment_builder
from orion.core.utils.format_trials import tuple_to_trial

log = logging.getLogger(__name__)

SHORT_DESCRIPTION = 'Inserts new trials in an existing experiment'
DESCRIPTION = 'Insert new trials for a given experiment with fixed values'


def add_subparser(parser):
    """Add the subparser that needs to be used for this command"""
    insert_parser = parser.add_parser('insert', help=SHORT_DESCRIPTION, description=DESCRIPTION)

    cli.get_basic_args_group(insert_parser)

    cli.get_user_args_group(insert_parser)

    insert_parser.set_defaults(func=main)
    insert_parser.set_defaults(help_empty=True)  # Print help if command is empty

    return insert_parser


def main(args):
    """Fetch config and insert new point"""
    command_line_user_args = args.pop('user_args', [None])[1:]
    experiment_view = experiment_builder.build_view_from_args(args)
    experiment = experiment_builder.build(name=experiment_view.name,
                                          version=experiment_view.version)

    transformed_args = _build_from(command_line_user_args)
    exp_space = experiment.space

    values = _create_tuple_from_values(transformed_args, exp_space)

    trial = tuple_to_trial(values, exp_space)

    experiment.register_trial(trial)


def _validate_dimensions(transformed_args, exp_space):
    exp_namespaces = list(exp_space.keys())
    values = {}

    # Find any dimension that is not given by the user and make sure they have a default value
    for exp_n in exp_namespaces:
        if exp_n not in transformed_args.keys() and exp_space[exp_n].default_value is None:
            error_msg = "Dimension {} is unspecified and has no default value".format(exp_n)
            raise ValueError(error_msg)

    # Find any namespace that is not in the space of the experiment,
    # or values that lie outside the prior's interval
    for namespace, value in transformed_args.items():
        if namespace not in exp_space:
            error_msg = "Found namespace outside of experiment space : {}".format(namespace)
            raise ValueError(error_msg)

        valid, value = _validate_input_value(value, exp_space, namespace)

        if not valid:
            error_msg = "Value {} is outside of dimension's prior interval".format(value)
            raise ValueError(error_msg)

        values[namespace] = value

    return values


def _validate_input_value(value, exp_space, namespace):

    is_valid = False

    try:
        value = eval(value)
        is_valid = value in exp_space[namespace]
    except NameError:
        is_valid = value in exp_space[namespace]

    return is_valid, value


def _create_tuple_from_values(transformed_args, exp_space):
    values_dict = _validate_dimensions(transformed_args, exp_space)
    values = []

    for namespace in exp_space:
        if namespace in values_dict.keys():
            values.append(values_dict[namespace])
        else:
            values.append(exp_space[namespace].default_value)

    return values


def _build_from(cmd_args):
    transformed_args, userconfig, _ = _build_from_args(cmd_args)

    if userconfig:
        transformed_args = _build_from_config(userconfig)

    return transformed_args


def _build_from_config(config_path):
    converter = infer_converter_from_file_type(config_path)
    userconfig_tmpl = converter.parse(config_path)
    userconfig_keyword = 'orion='

    transformed_args = {}
    stack = collections.deque()
    stack.append(('', userconfig_tmpl))
    while True:
        try:
            namespace, stuff = stack.pop()
        except IndexError:
            break
        if isinstance(stuff, dict):
            for k, v in stuff.items():
                stack.append(('/'.join([namespace, str(k)]), v))
        elif isinstance(stuff, list):
            for position, thing in enumerate(stuff):
                stack.append(('/'.join([namespace, str(position)]), thing))
        elif isinstance(stuff, str):
            if stuff.startswith(userconfig_keyword):
                if namespace in transformed_args:
                    error_msg = "Conflict for name '{}' in script configuration "\
                                "and arguments.".format(namespace)
                    raise ValueError(error_msg)

                transformed_args[namespace] = stuff[len(userconfig_keyword):]

    return transformed_args


def _build_from_args(cmd_args):
    userconfig = None
    is_userconfig_an_option = None
    userargs_search = r'\W*([a-zA-Z0-9_-]+)=([^ ]+)'
    userargs_regex_tmpl = r'(.*)=(.*)'
    userconfig_option = "--config="

    transformed_args = {}
    userargs_tmpl = collections.defaultdict(list)
    args_pattern = re.compile(userargs_search)
    args_prefix_pattern = re.compile(userargs_regex_tmpl)

    for arg in cmd_args:
        found = args_pattern.findall(arg)
        if len(found) != 1:
            if arg.startswith(userconfig_option):
                if not userconfig:
                    userconfig = arg[len(userconfig_option):]
                    is_userconfig_an_option = True
                else:
                    raise ValueError(
                        "Already found one configuration file in: %s" %
                        userconfig
                        )
            else:
                userargs_tmpl[None].append(arg)
            continue

        name, value = found[0]
        namespace = '/' + name

        if namespace in transformed_args:
            error_msg = "Conflict for name '{}' in script configuration "\
                        "and arguments.".format(namespace)
            raise ValueError(error_msg)

        transformed_args[namespace] = value

        found = args_prefix_pattern.findall(arg)
        assert len(found) == 1 and found[0][1] == value, "Parsing prefix problem."
        userargs_tmpl[namespace] = found[0][0] + '='

    if not userconfig and userargs_tmpl[None]:  # try the first positional argument
        if os.path.isfile(userargs_tmpl[None][0]):
            userconfig = userargs_tmpl[None].pop(0)
            is_userconfig_an_option = False

    return transformed_args, userconfig, is_userconfig_an_option



# -*- coding: utf-8 -*-
"""Tests :meth:`orion.analysis.regret()"""
import pandas as pd
import pytest

from orion.analysis.regret import regret


def test_accept_dict():
    """Tests dictionary parameter is supported"""
    data = {
        'id': ['a', 'b', 'c', 'd'],
        'objective': [0.1, 0.2, 0.3, 0.5]
    }

    result = regret(data)

    assert type(result) is pd.DataFrame


def test_accept_empty():
    """Tests an empty dataframe is returned if you give an empty dataframe"""
    empty_frame = pd.DataFrame()
    result = regret(empty_frame)

    assert result.empty
    assert result.equals(empty_frame)

    empty_frame = pd.DataFrame(columns=['id', 'objective'])
    result = regret(empty_frame)

    assert result.empty
    assert 'best' not in result.columns
    assert 'best_id' not in result.columns
    assert result.equals(empty_frame)


def test_parameter_not_modified():
    """Tests the original dataframe is not modified"""
    data = pd.DataFrame(data={
        'id': ['a', 'b', 'c', 'd'],
        'objective': [0.1, 0.2, 0.3, 0.5]
    })

    result = regret(data)

    assert len(data.columns) == 2
    assert len(result.columns) == 4


def test_length_names():
    """Tests the length of `names` parameter is two"""
    data = pd.DataFrame(data={
        'id': ['a', 'b'],
        'objective': [0.1, 0.2]
    })

    with pytest.raises(ValueError) as exception:
        regret(data, names=())
    assert "`names` requires a tuple with 2 elements. 0 provided." in str(exception.value)

    with pytest.raises(ValueError) as exception:
        regret(data, names=('a', 'b', 'c'))
    assert "`names` requires a tuple with 2 elements. 3 provided." in str(exception.value)


def test_default_column_names():
    """Tests default column names"""
    data = pd.DataFrame(data={
        'id': ['a', 'b', 'c', 'd'],
        'objective': [0.1, 0.2, 0.3, 0.5]
    })

    result = regret(data)

    assert 'best' in result.columns
    assert 'best_id' in result.columns


def test_input_column_names():
    """Tests custom column names"""
    data = pd.DataFrame(data={
        'id': ['a', 'b', 'c', 'd'],
        'objective': [0.1, 0.2, 0.3, 0.5]
    })

    result = regret(data, names=('something', 'else'))

    assert 'something' in result.columns
    assert 'else' in result.columns


def test_regret_sequential():
    """Tests that cumulative best objectives are linked to their respective ids"""
    data = pd.DataFrame(data={
        'id': ['a', 'b', 'c', 'd'],
        'objective': [10, 12, 8, 9],
    })

    expected_best = [10, 10, 8, 8]
    expected_ids = ['a', 'a', 'c', 'c']

    result = regret(data)

    assert all(result['best'].values == expected_best)
    assert all(result['best_id'].values == expected_ids)

    data = pd.DataFrame(data={
        'id': ['a', 'b', 'c'],
        'objective': [10, 9, 8],
    })

    expected_best = [10, 9, 8]
    expected_ids = ['a', 'b', 'c']

    result = regret(data)

    assert all(result['best'].values == expected_best)
    assert all(result['best_id'].values == expected_ids)

    data = pd.DataFrame(data={
        'id': ['a', 'b', 'c'],
        'objective': [8, 9, 10],
    })

    expected_best = [8, 8, 8]
    expected_ids = ['a', 'a', 'a']

    result = regret(data)

    assert all(result['best'].values == expected_best)
    assert all(result['best_id'].values == expected_ids)

    data = pd.DataFrame(data={
        'id': ['a'],
        'objective': [10],
    })

    result = regret(data)

    assert all(result['best'].values == [10])
    assert all(result['best_id'].values == ['a'])


def test_regret_equal():
    """Tests instances where trials share best objectives"""
    data = pd.DataFrame(data={
        'id': ['a', 'b', 'c'],
        'objective': [8, 9, 8],
    })

    expected_best = [8, 8, 8]
    expected_ids = ['a', 'a', 'a']

    result = regret(data)

    assert all(result['best'].values == expected_best)
    assert all(result['best_id'].values == expected_ids)

    data = pd.DataFrame(data={
        'id': ['a', 'b', 'c'],
        'objective': [8, 8, 8],
    })

    expected_best = [8, 8, 8]
    expected_ids = ['a', 'a', 'a']

    result = regret(data)

    assert all(result['best'].values == expected_best)
    assert all(result['best_id'].values == expected_ids)



# -*- coding: utf-8 -*-
"""
Oríon is an asynchronous distributed framework for black-box function optimization.

Its purpose is to serve as a hyperparameter optimizer for
machine learning models and training, as well as a flexible experimentation
platform for large scale asynchronous optimization procedures.

It has been designed firstly to disrupt a user's workflow at minimum, allowing
fast and efficient hyperparameter tuning, and secondly to provide secondary APIs
for more advanced features, such as dynamically reporting validation scores on
training time for automatic early stopping or on-the-fly reconfiguration.

Start by having a look here: https://github.com/epistimio/orion
"""
import logging
import os
import socket

from appdirs import AppDirs

from orion.core.io.config import Configuration
from ._version import get_versions


logger = logging.getLogger(__name__)


VERSIONS = get_versions()
del get_versions

__descr__ = 'Asynchronous [black-box] Optimization'
__version__ = VERSIONS['version']
__license__ = 'BSD-3-Clause'
__author__ = u'Epistímio'
__author_short__ = u'Epistímio'
__author_email__ = 'xavier.bouthillier@umontreal.ca'
__copyright__ = u'2017-2020, Epistímio'
__url__ = 'https://github.com/epistimio/orion'

DIRS = AppDirs(__name__, __author_short__)
del AppDirs

DEF_CONFIG_FILES_PATHS = [
    os.path.join(DIRS.site_data_dir, 'orion_config.yaml.example'),
    os.path.join(DIRS.site_config_dir, 'orion_config.yaml'),
    os.path.join(DIRS.user_config_dir, 'orion_config.yaml')
    ]


def define_config():
    """Create and define the fields of the configuration object."""
    config = Configuration()
    define_storage_config(config)
    define_experiment_config(config)
    define_worker_config(config)
    define_evc_config(config)

    config.add_option(
        'user_script_config', option_type=str, default='config',
        deprecate=dict(version='v0.3', alternative='worker.user_script_config'))

    config.add_option(
        'debug', option_type=bool, default=False,
        help='Turn Oríon into debug mode. Storage will be overriden to in-memory EphemeralDB.')

    return config


def define_storage_config(config):
    """Create and define the fields of the storage configuration."""
    storage_config = Configuration()

    storage_config.add_option(
        'type', option_type=str, default='legacy', env_var='ORION_STORAGE_TYPE')

    config.storage = storage_config

    define_database_config(config.storage)
    # Backward compatibility, should be removed in v0.3.0, or not?
    config.database = config.storage.database


def define_database_config(config):
    """Create and define the fields of the database configuration."""
    database_config = Configuration()

    try:
        default_host = socket.gethostbyname(socket.gethostname())
    except socket.gaierror:
        default_host = 'localhost'

    database_config.add_option(
        'name', option_type=str, default='orion', env_var='ORION_DB_NAME',
        help='Name of the database.')
    database_config.add_option(
        'type', option_type=str, default='MongoDB', env_var='ORION_DB_TYPE',
        help=('Type of database. Builtin backends are ``mongodb``, '
              '``pickleddb`` and ``ephemeraldb``.'))
    database_config.add_option(
        'host', option_type=str, default=default_host, env_var='ORION_DB_ADDRESS',
        help='URI for ``mongodb``, or file path for ``pickleddb``.')
    database_config.add_option(
        'port', option_type=int, default=27017, env_var='ORION_DB_PORT',
        help='Port address for ``mongodb``.')

    config.database = database_config


def define_experiment_config(config):
    """Create and define the fields of generic experiment configuration."""
    experiment_config = Configuration()

    experiment_config.add_option(
        'max_trials', option_type=int, default=int(10e8), env_var='ORION_EXP_MAX_TRIALS',
        help="number of trials to be completed for the experiment. This value "
             "will be saved within the experiment configuration and reused "
             "across all workers to determine experiment's completion. ")

    experiment_config.add_option(
        'worker_trials', option_type=int, default=int(10e8),
        deprecate=dict(version='v0.3', alternative='worker.max_trials',
                       name='experiment.worker_trials'),
        help="This argument will be removed in v0.3. Use --worker-max-trials instead.")

    experiment_config.add_option(
        'max_broken', option_type=int, default=3, env_var='ORION_EXP_MAX_BROKEN',
        help=('Maximum number of broken trials before experiment stops.'))

    experiment_config.add_option(
        'working_dir', option_type=str, default='', env_var='ORION_WORKING_DIR',
        help="Set working directory for running experiment.")

    experiment_config.add_option(
        "pool_size", option_type=int, default=1,
        deprecate=dict(version='v0.3', alternative=None, name='experiment.pool_size'),
        help="This argument will be removed in v0.3.")

    experiment_config.add_option(
        'algorithms', option_type=dict, default={'random': {'seed': None}},
        help='Algorithm configuration for the experiment.')

    experiment_config.add_option(
        'strategy', option_type=dict, default={'MaxParallelStrategy': {}},
        help='Parallel strategy to use with the algorithm.')

    config.experiment = experiment_config


def define_worker_config(config):
    """Create and define the fields of the worker configuration."""
    worker_config = Configuration()

    worker_config.add_option(
        'heartbeat', option_type=int, default=120, env_var='ORION_HEARTBEAT',
        help=('Frequency (seconds) at which the heartbeat of the trial is updated. '
              'If the heartbeat of a `reserved` trial is larger than twice the configured '
              'heartbeat, Oríon will reset the status of the trial to `interrupted`. '
              'This allows restoring lost trials (ex: due to killed worker).'))

    worker_config.add_option(
        'max_trials', option_type=int, default=int(10e8), env_var='ORION_WORKER_MAX_TRIALS',
        help="number of trials to be completed for this worker. "
             "If the experiment is completed, the worker will die even if it "
             "did not reach its maximum number of trials ")

    worker_config.add_option(
        'max_broken', option_type=int, default=3, env_var='ORION_WORKER_MAX_BROKEN',
        help=('Maximum number of broken trials before worker stops.'))

    worker_config.add_option(
        'max_idle_time', option_type=int, default=60, env_var='ORION_MAX_IDLE_TIME',
        help=('Maximum time the producer can spend trying to generate a new suggestion.'
              'Such timeout are generally caused by slow database, large number of '
              'concurrent workers leading to many race conditions or small search spaces '
              'with integer/categorical dimensions that may be fully explored.'))

    worker_config.add_option(
        'interrupt_signal_code', option_type=int, default=130, env_var='ORION_INTERRUPT_CODE',
        help='Signal returned by user script to signal to Oríon that it was interrupted.')

    # TODO: Will this support -config as well, or only --config?
    worker_config.add_option(
        'user_script_config', option_type=str, default='config',
        env_var='ORION_USER_SCRIPT_CONFIG',
        help='Config argument name of user\'s script (--config).')

    config.worker = worker_config


def define_evc_config(config):
    """Create and define the fields of the evc configuration."""
    evc_config = Configuration()

    # TODO: This should be built automatically like get_branching_args_group
    #       After this, the cmdline parser should be built based on config.

    evc_config.add_option(
        'auto_resolution', option_type=bool, default=True,
        deprecate=dict(version='v0.3', alternative='evc.manual_resolution',
                       name='evc.auto_resolution'),
        help="This argument will be removed in v0.3. "
             "Conflicts are now resolved automatically by default. "
             "See --manual-resolution to avoid auto-resolution.")

    evc_config.add_option(
        'manual_resolution', option_type=bool, default=False,
        env_var='ORION_EVC_MANUAL_RESOLUTION',
        help=("If ``True``, enter experiment version control conflict resolver for "
              "manual resolution on branching events. Otherwise, auto-resolution is "
              "attempted."))

    evc_config.add_option(
        'non_monitored_arguments', option_type=list, default=[],
        env_var='ORION_EVC_NON_MONITORED_ARGUMENTS',
        help=("Ignore these commandline arguments when looking for differences in "
              "user's commandline call. "
              "Environment variable and commandline only supports one argument. "
              "Use global config or local config to pass a list of arguments to ignore."))

    evc_config.add_option(
        'ignore_code_changes', option_type=bool, default=False,
        env_var='ORION_EVC_IGNORE_CODE_CHANGES',
        help=("If ``True``, ignore code changes when looking for differences."))

    evc_config.add_option(
        'algorithm_change', option_type=bool, default=False,
        env_var='ORION_EVC_ALGO_CHANGE',
        help=("Set algorithm change as resolved if a branching event occur. "
              "Child and parent experiment have access to all trials from each other "
              "when the only difference between them is the algorithm configuration."))

    evc_config.add_option(
        'code_change_type', option_type=str, default='break',
        env_var='ORION_EVC_CODE_CHANGE',
        help=("One of ``break``, ``unsure`` or ``noeffet``. "
              "Defines how trials should be filtered in Experiment Version Control tree "
              "if there is a change in the user's code repository. "
              "If the effect of the change is ``unsure``, "
              "the child experiment will access the trials of the parent but not "
              "the other way around. "
              "This is to ensure parent experiment does not get corrupted with possibly "
              "incompatible results. "
              "The child cannot access the trials from parent if ``code_change_type`` "
              "is ``break``. The parent cannot access trials from child if "
              "``code_change_type`` is ``unsure`` or ``break``."))

    evc_config.add_option(
        'cli_change_type', option_type=str, default='break',
        env_var='ORION_EVC_CMDLINE_CHANGE',
        help=("One of ``break``, ``unsure`` or ``noeffet``. "
              "Defines how trials should be filtered in Experiment Version Control tree "
              "if there is a change in the user's commandline call. "
              "If the effect of the change is ``unsure``, "
              "the child experiment will access the trials of the parent but not "
              "the other way around. "
              "This is to ensure parent experiment does not get corrupted with possibly "
              "incompatible results. "
              "The child cannot access the trials from parent if ``cli_change_type`` "
              "is ``break``. The parent cannot access trials from child if "
              "``cli_change_type`` is ``unsure`` or ``break``."))

    evc_config.add_option(
        'config_change_type', option_type=str, default='break',
        env_var='ORION_EVC_CONFIG_CHANGE',
        help=("One of ``break``, ``unsure`` or ``noeffet``. "
              "Defines how trials should be filtered in Experiment Version Control tree "
              "if there is a change in the user's script. "
              "If the effect of the change is ``unsure``, "
              "the child experiment will access the trials of the parent but not "
              "the other way around. "
              "This is to ensure parent experiment does not get corrupted with possibly "
              "incompatible results. "
              "The child cannot access the trials from parent if ``config_change_type`` "
              "is ``break``. The parent cannot access trials from child if "
              "``config_change_type`` is ``unsure`` or ``break``."))

    config.evc = evc_config


def build_config():
    """Define the config and fill it based on global configuration files."""
    config = define_config()
    for file_path in DEF_CONFIG_FILES_PATHS:
        if not os.path.exists(file_path):
            logger.debug('Config file not found: %s', file_path)
            continue

        config.load_yaml(file_path)

    return config


config = build_config()



# -*- coding: utf-8 -*-
"""
:mod:`orion.serving.root_resource` -- Represent the '/' REST endpoint
=====================================================================================

.. module:: root_resource
   :platform: Unix
   :synopsis: Represent the '/' REST endpoint
"""
import json

import orion.core
from orion.storage.base import get_storage


class RuntimeResource(object):
    """Handle requests for the '/' REST endpoint"""

    def __init__(self):
        pass

    def on_get(self, req, resp):
        """Handle the HTTP GET requests for the '/' endpoint

        Parameters
        ----------
        req
            The request
        resp
            The response to send back
        """
        database = get_storage()._db.__class__.__name__
        response = {
            'orion': orion.core.__version__,
            'server': 'gunicorn',
            'database': database
        }

        resp.body = json.dumps(response, indent=4)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Collection of tests for :mod:`orion.core.worker.consumer`."""
import os
import signal
import subprocess
import time

import pytest

import orion.core.io.experiment_builder as experiment_builder
import orion.core.io.resolve_config as resolve_config
import orion.core.utils.backward as backward
from orion.core.utils.exceptions import BranchingEvent
from orion.core.utils.format_trials import tuple_to_trial
import orion.core.worker.consumer as consumer


Consumer = consumer.Consumer


@pytest.fixture
def config(exp_config):
    """Return a configuration."""
    config = exp_config[0][0]
    config['metadata']['user_args'] = ['--x~uniform(-50, 50)']
    config['metadata']['VCS'] = resolve_config.infer_versioning_metadata(
        config['metadata']['user_script'])
    config['name'] = 'exp'
    config['working_dir'] = "/tmp/orion"
    backward.populate_space(config)
    config['space'] = config['metadata']['priors']
    return config


@pytest.mark.usefixtures("create_db_instance")
def test_trials_interrupted_keyboard_int(config, monkeypatch):
    """Check if a trial is set as interrupted when a KeyboardInterrupt is raised."""
    def mock_Popen(*args, **kwargs):
        raise KeyboardInterrupt

    exp = experiment_builder.build(**config)

    monkeypatch.setattr(consumer.subprocess, "Popen", mock_Popen)

    trial = tuple_to_trial((1.0,), exp.space)

    exp.register_trial(trial)

    con = Consumer(exp)

    with pytest.raises(KeyboardInterrupt):
        con.consume(trial)

    trials = exp.fetch_trials_by_status('interrupted')
    assert len(trials)
    assert trials[0].id == trial.id


@pytest.mark.usefixtures("create_db_instance")
def test_trials_interrupted_sigterm(config, monkeypatch):
    """Check if a trial is set as interrupted when a signal is raised."""
    def mock_popen(*args, **kwargs):
        os.kill(os.getpid(), signal.SIGTERM)

    exp = experiment_builder.build(**config)

    monkeypatch.setattr(subprocess.Popen, "wait", mock_popen)

    trial = tuple_to_trial((1.0,), exp.space)

    exp.register_trial(trial)

    con = Consumer(exp)

    with pytest.raises(KeyboardInterrupt):
        con.consume(trial)

    trials = exp.fetch_trials_by_status('interrupted')
    assert len(trials)
    assert trials[0].id == trial.id


@pytest.mark.usefixtures("create_db_instance")
def test_pacemaker_termination(config):
    """Check if pacemaker stops as soon as the trial completes."""
    exp = experiment_builder.build(**config)

    trial = tuple_to_trial((1.0,), exp.space)

    exp.register_trial(trial, status='reserved')

    con = Consumer(exp)

    start = time.time()

    con.consume(trial)
    con.pacemaker.join()

    duration = time.time() - start

    assert duration < con.pacemaker.wait_time


@pytest.mark.usefixtures("create_db_instance")
def test_trial_working_dir_is_changed(config):
    """Check that trial has its working_dir attribute changed."""
    exp = experiment_builder.build(**config)

    trial = tuple_to_trial((1.0,), exp.space)

    exp.register_trial(trial, status='reserved')

    con = Consumer(exp)
    con.consume(trial)

    assert trial.working_dir is not None
    assert trial.working_dir == con.working_dir + "/exp_" + trial.id


@pytest.mark.usefixtures("create_db_instance")
def test_code_changed(config, monkeypatch):
    """Check that trial has its working_dir attribute changed."""
    exp = experiment_builder.build(**config)

    trial = tuple_to_trial((1.0,), exp.space)

    exp.register_trial(trial, status='reserved')

    con = Consumer(exp)

    def code_changed(user_script):
        return dict(
            type='git',
            is_dirty=True,
            HEAD_sha='changed',
            active_branch='new_branch',
            diff_sha='new_diff')

    monkeypatch.setattr(consumer, 'infer_versioning_metadata', code_changed)

    with pytest.raises(BranchingEvent) as exc:
        con.consume(trial)

    assert exc.match('Code changed between execution of 2 trials')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Perform functional tests for the parsing of the `hunt` command."""
import argparse
import os

import pytest

from orion.core.cli import hunt


def _create_parser(need_subparser=True):
    parser = argparse.ArgumentParser()

    if need_subparser:
        subparsers = parser.add_subparsers()
        return parser, subparsers

    return parser


@pytest.mark.usefixtures("clean_db")
def test_hunt_command_full_parsing(database, monkeypatch):
    """Test the parsing of the `hunt` command"""
    monkeypatch.chdir(os.path.dirname(os.path.abspath(__file__)))
    parser, subparsers = _create_parser()
    args_list = ["hunt", "-n", "test",
                 "--config", "./orion_config_random.yaml",
                 "--max-trials", "400", "--pool-size", "4", "--worker-trials", "5",
                 "./black_box.py", "-x~normal(1,1)"]

    hunt.add_subparser(subparsers)
    subparsers.choices['hunt'].set_defaults(func='')

    args = vars(parser.parse_args(args_list))
    assert args['name'] == 'test'
    assert args['config'].name == './orion_config_random.yaml'
    assert args['user_args'] == ['./black_box.py', '-x~normal(1,1)']
    assert args['pool_size'] == 4
    assert args['max_trials'] == 400
    assert args['worker_trials'] == 5



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.list` -- Module to list experiments
========================================================
.. module:: list
   :platform: Unix
   :synopsis: List experiments in terminal
"""
import logging

from orion.core.cli import base as cli
import orion.core.io.experiment_builder as experiment_builder
from orion.core.utils.pptree import print_tree
from orion.storage.base import get_storage

log = logging.getLogger(__name__)
SHORT_DESCRIPTION = 'Gives a list of experiments'
DESCRIPTION = """
This command gives a list of your experiments in a easy-to-read tree-like structure.
"""


def add_subparser(parser):
    """Add the subparser that needs to be used for this command"""
    list_parser = parser.add_parser('list', help=SHORT_DESCRIPTION, description=DESCRIPTION)

    cli.get_basic_args_group(list_parser)

    list_parser.set_defaults(func=main)

    return list_parser


def main(args):
    """List all experiments inside database."""
    config = experiment_builder.get_cmd_config(args)
    experiment_builder.setup_storage(config.get('storage'))

    query = {}

    if args['name']:
        query['name'] = args['name']

    experiments = get_storage().fetch_experiments(query)

    if args['name']:
        root_experiments = experiments
    else:
        root_experiments = [exp for exp in experiments
                            if exp['refers'].get('root_id', exp['_id']) == exp['_id']]

    if not root_experiments:
        print("No experiment found")
        return

    for root_experiment in root_experiments:
        root = experiment_builder.build_view(name=root_experiment['name'],
                                             version=root_experiment.get('version')).node
        print_tree(root, nameattr='tree_name')



# -*- coding: utf-8 -*-
"""
:mod:`orion.client.manual` -- Routines for manual interaction with database
=============================================================================

.. module:: manual
   :platform: Unix
   :synopsis: Provides a simple interface to log new trials into the database
      and link them with a particular existing experiment.

"""
import logging

from orion.client import create_experiment
from orion.core.utils import format_trials


log = logging.getLogger(__name__)


def insert_trials(experiment_name, points, raise_exc=True):
    """Insert sets of parameters manually, defined in `points`, as new trials
    for the experiment name, `experiment_name`.

    .. warning::

        This function is deprecated and will be removed in 0.3.0.
        You should use ExperimentClient.insert() instead.

    :param experiment_name: Name of the experiment which the new trials are
       going to be associated with
    :param points: list of tuples in agreement with experiment's parameter space
    :param raise_exc: whether an inappropriate tuple of parameters will raise
       an exception or it will be ignored

    .. note:: If `raise_exc` is True, no set of parameters will be inserted. If
       it is False, only the valid ones will be inserted; the rest will be ignored.

    .. note:: This cannot be used to prepopulate a future experiment. So,
       an experiment with `experiment_name` should already be configured in
       the database.

    """
    log.warning('insert_trials() is deprecated and will be removed in 0.3.0. '
                'You should use ExperimentClient.insert() instead.')
    experiment = create_experiment(experiment_name)

    valid_points = []

    for point in points:
        try:
            assert point in experiment.space
            valid_points.append(point)
        except AssertionError:
            if raise_exc:
                raise

    if not valid_points:
        return

    new_trials = list(
        map(lambda data: format_trials.tuple_to_trial(data, experiment.space),
            valid_points))

    for new_trial in new_trials:
        experiment.insert(new_trial.params)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.init_only` -- Module running the init_only command
=======================================================================

.. module:: init_only
   :platform: Unix
   :synopsis: Creates a new experiment.
"""

import logging

import orion.core
from orion.core.cli import base as cli
from orion.core.cli import evc as evc_cli
import orion.core.io.experiment_builder as experiment_builder

log = logging.getLogger(__name__)
DESCRIPTION = '(DEPRECATED) Use command `orion hunt --init_only` instead'


def add_subparser(parser):
    """Return the parser that needs to be used for this command"""
    init_only_parser = parser.add_parser('init_only', help=DESCRIPTION, description=DESCRIPTION)

    orion_group = cli.get_basic_args_group(
        init_only_parser, group_name='init_only arguments', group_help='')

    orion.core.config.experiment.add_arguments(
        orion_group,
        rename=dict(max_broken='--exp-max-broken', max_trials='--exp-max-trials'))

    orion_group.add_argument(
        '--max-trials', type=int, metavar='#',
        help="(DEPRECATED) This argument will be removed in v0.3. Use --exp-max-trials instead")

    evc_cli.get_branching_args_group(init_only_parser)

    cli.get_user_args_group(init_only_parser)

    init_only_parser.set_defaults(func=main)
    init_only_parser.set_defaults(help_empty=True)  # Print help if command is empty

    return init_only_parser


def main(args):
    """Build and initialize experiment"""
    # By building the experiment, we create a new experiment document in database
    log.warning('Command init_only is deprecated and will be removed in v0.3. '
                'Use orion hunt --init-only instead.')
    experiment_builder.build_from_args(args)



# -*- coding: utf-8 -*-
"""
:mod:`orion.algo.random` -- Random sampler as optimization algorithm
======================================================================

.. module:: random
   :platform: Unix
   :synopsis: Draw and deliver samples from prior defined in problem's domain.

"""
import numpy

from orion.algo.base import BaseAlgorithm, infer_trial_id


class Random(BaseAlgorithm):
    """Implement a algorithm that samples randomly from the problem's space."""

    def __init__(self, space, seed=None):
        """Random sampler takes no other hyperparameter than the problem's space
        itself.

        :param space: `orion.algo.space.Space` of optimization.
        :param seed: Integer seed for the random number generator.
        """
        super(Random, self).__init__(space, seed=seed)

    def seed_rng(self, seed):
        """Seed the state of the random number generator.

        :param seed: Integer seed for the random number generator.
        """
        self.rng = numpy.random.RandomState(seed)

    @property
    def state_dict(self):
        """Return a state dict that can be used to reset the state of the algorithm."""
        _state_dict = super(Random, self).state_dict
        _state_dict['rng_state'] = self.rng.get_state()
        return _state_dict

    def set_state(self, state_dict):
        """Reset the state of the algorithm based on the given state_dict

        :param state_dict: Dictionary representing state of an algorithm
        """
        super(Random, self).set_state(state_dict)
        self.seed_rng(0)
        self.rng.set_state(state_dict['rng_state'])

    def suggest(self, num=1):
        """Suggest a `num` of new sets of parameters. Randomly draw samples
        from the import space and return them.

        :param num: how many sets to be suggested.

        .. note:: New parameters must be compliant with the problem's domain
           `orion.algo.space.Space`.
        """
        points = []
        point_ids = set(self._trials_info.keys())
        i = 0
        while len(points) < num:
            new_point = self.space.sample(1, seed=tuple(self.rng.randint(0, 1000000, size=3)))[0]
            point_id = infer_trial_id(new_point)
            if point_id not in point_ids:
                point_ids.add(point_id)
                points.append(new_point)
            i += 1

        return points



# -*- coding: utf-8 -*-
"""
:mod:`orion.core.cli.evc` -- Function utilities for evc in cli
==============================================================

.. module:: evc
   :platform: Unix
   :synopsis: Orion helper functions to parse command-line options for branching resolution

"""
from orion.core.evc import adapters
from orion.core.evc.conflicts import Resolution


def _add_auto_resolution_argument(parser):
    parser.add_argument(
        "--auto-resolution",
        help="Deprecated. Conflicts are now resolved automatically by default."
             "See --manual-resolution to avoid auto-resolution.")


def _add_manual_resolution_argument(parser):
    parser.add_argument(
        "--manual-resolution",
        action="store_true",
        help="Manually resolve conflicts")


def _add_non_monitored_arguments_argument(parser):
    parser.add_argument(
        "--non-monitored-arguments", type=str, nargs='*',
        help="Ignore these arguments when looking for differences")


def _add_ignore_code_changes_argument(parser):
    parser.add_argument(
        "--ignore-code-changes",
        action="store_true",
        help="Ignore code changes when looking for differences")


def _add_branch_from_argument(parser):
    parser.add_argument(
        "--branch-from", type=str,
        help="Create a new child based on experiment defined by `branch-from`")


def _add_algorithm_argument(parser, resolution_class):
    parser.add_argument(
        resolution_class.ARGUMENT,
        action="store_true",
        help="Set algorithm change as resolved if a branching event occur")


def _add_code_argument(parser, resolution_class):
    parser.add_argument(
        resolution_class.ARGUMENT,
        choices=adapters.CodeChange.types,
        help="Set code change type")


def _add_cli_argument(parser, resolution_class):
    parser.add_argument(
        resolution_class.ARGUMENT,
        choices=adapters.CommandLineChange.types,
        help="Set command line change type")


def _add_config_argument(parser, resolution_class):
    parser.add_argument(
        resolution_class.ARGUMENT,
        choices=adapters.ScriptConfigChange.types,
        help="Set configuration change type")


def _add_branch_to_argument(parser, resolution_class):
    parser.add_argument(
        '-b', resolution_class.ARGUMENT, metavar='stringID',
        help='Unique name for the new branching experiment')


resolution_arguments = {
    'auto_resolution': _add_auto_resolution_argument,
    'manual_resolution': _add_manual_resolution_argument,
    'non_monitored_arguments': _add_non_monitored_arguments_argument,
    'ignore_code_changes': _add_ignore_code_changes_argument,
    'algorithm_change': _add_algorithm_argument,
    'code_change_type': _add_code_argument,
    'cli_change_type': _add_cli_argument,
    'config_change_type': _add_config_argument,
    'branch_from': _add_branch_from_argument,
    'branch_to': _add_branch_to_argument}


UNDEFINED_PARSER_ERROR = (
    "A resolution with metavar '{}' is defined but no corresponding parser is defined. "
    "Please raise an issue on github if you encounter this error message.")


def get_branching_args_group(parser):
    """Return the arguments for automatic branching resolution."""
    branching_args_group = parser.add_argument_group(
        "Oríon branching arguments (optional)",
        description="Arguments to automatically resolved branching events.")

    _add_manual_resolution_argument(branching_args_group)
    _add_non_monitored_arguments_argument(branching_args_group)
    _add_ignore_code_changes_argument(branching_args_group)
    _add_auto_resolution_argument(branching_args_group)
    _add_branch_from_argument(branching_args_group)

    for resolution_class in sorted(Resolution.__subclasses__(), key=lambda cls: cls.__name__):
        if not resolution_class.ARGUMENT:
            continue

        metavar = resolution_class.namespace()

        # This should fail blatantly if we forget to create a parser for a new resolution which
        # needs argument in command-line for automatic resolution.
        assert metavar in resolution_arguments, UNDEFINED_PARSER_ERROR.format(metavar)

        resolution_arguments[metavar](branching_args_group, resolution_class)

    return branching_args_group


def fetch_branching_configuration(config):
    """Build a dictionary of arguments for branching"""
    return {key: config[key] for key in resolution_arguments if key in config}
